# ======================================================================
# Copyright 2020 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

from unittest.mock import call, Mock

import pytest

from supvisors.statscollector import LocalNodeInfo
from supvisors.web.viewcontext import ViewContext
from supvisors.web.viewhostinstance import HostInstanceView
from supvisors.web.viewimage import (host_cpu_img, host_net_io_img, host_mem_img,
                                     host_disk_io_img, host_disk_usage_img)
from supvisors.web.webutils import HOST_INSTANCE_PAGE, PROC_INSTANCE_PAGE
from .base import DummyHttpContext
from .conftest import create_element


@pytest.fixture
def http_context(supvisors):
    """ Fixture for a consistent mocked HTTP context provided by Supervisor. """
    http_context = DummyHttpContext('ui/host_instance.html')
    http_context.supervisord.supvisors = supvisors
    supvisors.supervisor_data.supervisord = http_context.supervisord
    return http_context


@pytest.fixture
def view(http_context):
    """ Fixture for the instance to test. """
    # create the instance to be tested
    return HostInstanceView(http_context)


def test_init(view):
    """ Test the values set at construction. """
    assert view.page_name == HOST_INSTANCE_PAGE


def test_write_options(mocker, view):
    """ Test the ApplicationView.write_options method. """
    mocked_period = mocker.patch.object(view, 'write_periods')
    mocked_switch = mocker.patch.object(view, 'write_view_switch')
    mocked_header = Mock()
    view.write_options(mocked_header)
    assert mocked_period.call_args_list == [call(mocked_header)]
    assert mocked_switch.call_args_list == [call(mocked_header)]


def test_write_view_switch(supvisors, view):
    """ Test the SupvisorsInstanceView.write_view_switch method. """
    # set context (meant to be set through constructor and render)
    view.view_ctx = Mock(**{'format_url.return_value': 'an url'})
    supvisors.mapper.local_identifier = '10.0.0.1:25000'
    # build root structure
    mocked_process_view_mid = create_element()
    mocked_host_view_mid = create_element()
    mocked_header = create_element({'process_view_a_mid': mocked_process_view_mid,
                                    'host_view_a_mid': mocked_host_view_mid})
    # test call when SupvisorsInstanceView is a host page
    view.page_name = HOST_INSTANCE_PAGE
    view.write_view_switch(mocked_header)
    assert mocked_header.findmeld.call_args_list == [call('process_view_a_mid'), call('host_view_a_mid'), ]
    assert view.view_ctx.format_url.call_args_list == [call('', PROC_INSTANCE_PAGE)]
    assert mocked_process_view_mid.attributes.call_args_list == [call(href='an url')]
    assert mocked_host_view_mid.content.call_args_list == [call('10.0.0.1')]


def test_write_contents_no_plot(mocker, supvisors, view):
    """ Test the write_contents method in the context where psutil is not installed. """
    mocked_characteristics = mocker.patch.object(view, 'write_node_characteristics')
    mocked_processor = mocker.patch.object(view, 'write_processor_statistics')
    mocked_memory = mocker.patch.object(view, 'write_memory_statistics')
    mocked_network = mocker.patch.object(view, 'write_network_io_statistics')
    mocked_disk = mocker.patch.object(view, 'write_disk_statistics')
    mocked_cpu_img = mocker.patch.object(view, '_write_cpu_image', side_effect=ImportError)
    mocked_mem_img = mocker.patch.object(view, '_write_mem_image')
    mocked_net_io_img = mocker.patch.object(view, '_write_net_io_image')
    mocked_disk_img = mocker.patch.object(view, '_write_disk_image')
    # set context (meant to be set through render)
    dummy_stats = Mock(cpu='cpu', mem='mem', net_io='net_io', times='times')
    view.view_ctx = Mock(**{'get_instance_stats.return_value': dummy_stats,
                            'get_node_characteristics.return_value': supvisors.stats_collector.node_info})
    # create xhtml structure
    cpu_image_fig_mid = create_element()
    mem_image_fig_mid = create_element()
    net_io_image_fig_mid = create_element()
    disk_io_image_fig_mid = create_element()
    disk_usage_image_fig_mid = create_element()
    contents_elt = create_element({'cpu_image_fig_mid': cpu_image_fig_mid,
                                   'mem_image_fig_mid': mem_image_fig_mid,
                                   'net_io_image_fig_mid': net_io_image_fig_mid,
                                   'disk_io_image_fig_mid': disk_io_image_fig_mid,
                                   'disk_usage_image_fig_mid': disk_usage_image_fig_mid})
    # test call
    view.write_contents(contents_elt)
    assert mocked_characteristics.call_args_list == [call(contents_elt, supvisors.stats_collector.node_info)]
    assert mocked_processor.call_args_list == [call(contents_elt, 'cpu', 'times')]
    assert mocked_memory.call_args_list == [call(contents_elt, 'mem', 'times')]
    assert mocked_network.call_args_list == [call(contents_elt, 'net_io')]
    assert mocked_disk.call_args_list == [call(contents_elt, dummy_stats)]
    assert mocked_cpu_img.call_args_list == [call('cpu', 'times')]
    assert not mocked_mem_img.called
    assert not mocked_net_io_img.called
    assert not mocked_disk_img.called
    assert contents_elt.findmeld.call_args_list == [call('cpu_image_fig_mid'),
                                                    call('mem_image_fig_mid'),
                                                    call('net_io_image_fig_mid'),
                                                    call('disk_io_image_fig_mid'),
                                                    call('disk_usage_image_fig_mid')]
    assert cpu_image_fig_mid.replace.call_args_list == [call('')]
    assert mem_image_fig_mid.replace.call_args_list == [call('')]
    assert net_io_image_fig_mid.replace.call_args_list == [call('')]
    assert disk_io_image_fig_mid.replace.call_args_list == [call('')]
    assert disk_usage_image_fig_mid.replace.call_args_list == [call('')]


def test_write_contents(mocker, supvisors, view):
    """ Test the write_contents method. """
    # skip test if matplotlib is not installed
    pytest.importorskip('matplotlib', reason='cannot test as optional matplotlib is not installed')
    mocked_characteristics = mocker.patch.object(view, 'write_node_characteristics')
    mocked_processor = mocker.patch.object(view, 'write_processor_statistics')
    mocked_memory = mocker.patch.object(view, 'write_memory_statistics')
    mocked_network = mocker.patch.object(view, 'write_network_io_statistics')
    mocked_disk = mocker.patch.object(view, 'write_disk_statistics')
    mocked_cpu_img = mocker.patch.object(view, '_write_cpu_image')
    mocked_mem_img = mocker.patch.object(view, '_write_mem_image')
    mocked_net_io_img = mocker.patch.object(view, '_write_net_io_image')
    mocked_disk_img = mocker.patch.object(view, '_write_disk_image')
    # set context (meant to be set through render)
    dummy_stats = Mock(cpu='cpu', mem='mem', net_io='net_io', times='times')
    view.view_ctx = Mock(**{'get_instance_stats.return_value': dummy_stats,
                            'get_node_characteristics.return_value': supvisors.stats_collector.node_info})
    # create xhtml structure
    cpu_image_fig_mid = create_element()
    mem_image_fig_mid = create_element()
    net_io_image_fig_mid = create_element()
    disk_io_image_fig_mid = create_element()
    disk_usage_image_fig_mid = create_element()
    contents_elt = create_element({'cpu_image_fig_mid': cpu_image_fig_mid,
                                   'mem_image_fig_mid': mem_image_fig_mid,
                                   'net_io_image_fig_mid': net_io_image_fig_mid,
                                   'disk_io_image_fig_mid': disk_io_image_fig_mid,
                                   'disk_usage_image_fig_mid': disk_usage_image_fig_mid})
    # test call
    view.write_contents(contents_elt)
    assert mocked_characteristics.call_args_list == [call(contents_elt, supvisors.stats_collector.node_info)]
    assert mocked_processor.call_args_list == [call(contents_elt, 'cpu', 'times')]
    assert mocked_memory.call_args_list == [call(contents_elt, 'mem', 'times')]
    assert mocked_network.call_args_list == [call(contents_elt, 'net_io')]
    assert mocked_disk.call_args_list == [call(contents_elt, dummy_stats)]
    assert mocked_cpu_img.call_args_list == [call('cpu', 'times')]
    assert mocked_mem_img.call_args_list == [call('mem', 'times')]
    assert mocked_net_io_img.call_args_list == [call('net_io')]
    assert mocked_disk_img.call_args_list == [call(dummy_stats)]
    assert not contents_elt.findmeld.called
    assert not cpu_image_fig_mid.replace.called
    assert not mem_image_fig_mid.replace.called
    assert not net_io_image_fig_mid.replace.called
    assert not disk_io_image_fig_mid.replace.called
    assert not disk_usage_image_fig_mid.replace.called


def test_write_node_characteristics(mocker, supvisors, view):
    """ Test the write_node_characteristics method. """
    # patch psutil functions
    mocker.patch('psutil.cpu_count', return_value=4)
    mocker.patch('psutil.cpu_freq', return_value=Mock(current=3000.01))
    mocker.patch('psutil.virtual_memory', return_value=Mock(total=16123456789))
    # create the xhtml structure
    node_td_mid = create_element()
    ipaddress_td_mid = create_element()
    cpu_count_td_mid = create_element()
    cpu_freq_td_mid = create_element()
    physical_mem_td_mid = create_element()
    contents_elt = create_element({'node_td_mid': node_td_mid, 'ipaddress_td_mid': ipaddress_td_mid,
                                   'cpu_count_td_mid': cpu_count_td_mid, 'cpu_freq_td_mid': cpu_freq_td_mid,
                                   'physical_mem_td_mid': physical_mem_td_mid})
    # change local node
    supvisors.mapper.local_identifier = '10.0.0.1:25000'
    # test call
    node_info = LocalNodeInfo()
    view.write_node_characteristics(contents_elt, node_info)
    assert node_td_mid.content.call_args_list == [call('10.0.0.1')]
    assert ipaddress_td_mid.content.call_args_list == [call('10.0.0.1')]
    assert cpu_count_td_mid.content.call_args_list == [call('4 / 4')]
    assert cpu_freq_td_mid.content.call_args_list == [call('3000 MHz')]
    assert physical_mem_td_mid.content.call_args_list == [call('15.02 GiB')]


def test_write_processor_single_title(view):
    """ Test the _write_processor_single_title method. """
    # set context (meant to be set through render)
    view.view_ctx = Mock(**{'format_url.return_value': 'http://addr:port/index.html',
                            'cpu_id_to_string.return_value': '1'})  # replace root structure
    mocked_title_mid = Mock(attrib={})
    mocked_tr = Mock(**{'findmeld.return_value': mocked_title_mid})
    # in first call, elt is not the selected element
    view._write_processor_single_title(mocked_tr, 1, 0)
    assert mocked_tr.findmeld.call_args_list == [call('cpu_num_a_mid')]
    assert mocked_title_mid.attrib == {}
    assert mocked_title_mid.attributes.call_args_list == [call(href='http://addr:port/index.html')]
    assert mocked_title_mid.content.call_args_list == [call('cpu#1')]
    mocked_tr.findmeld.reset_mock()
    mocked_title_mid.attributes.reset_mock()
    # in first call, elt is the selected element
    view._write_processor_single_title(mocked_tr, 1, 1)
    assert mocked_tr.findmeld.call_args_list == [call('cpu_num_a_mid')]
    assert mocked_title_mid.attrib == {'class': 'button off active'}
    assert not mocked_title_mid.attributes.called
    assert mocked_title_mid.content.call_args_list == [call('cpu#1')]


def test_write_processor_single_statistics(mocker, view):
    """ Test the _write_processor_single_statistics method. """
    mocked_common = mocker.patch.object(view, '_write_common_detailed_statistics')
    # replace root element
    mocked_root = Mock()
    # test method call
    view._write_processor_single_statistics(mocked_root, [1.523, 2.456], [1, 2, 3])
    assert mocked_common.call_args_list == [call(mocked_root, [1.523, 2.456], [1, 2, 3],
                                                 'cpu_val_td_mid', 'cpu_avg_td_mid',
                                                 'cpu_slope_td_mid', 'cpu_dev_td_mid')]


def test_write_processor_statistics(mocker, view):
    """ Test the write_processor_statistics method. """
    mocked_stats = mocker.patch.object(view, '_write_processor_single_statistics')
    mocked_title = mocker.patch.object(view, '_write_processor_single_title')
    # set context (meant to be set through render)
    view.view_ctx = Mock(cpu_id=1)
    # build root structure
    mocked_trs = [Mock(attrib={}) for _ in range(2)]
    mocked_mid = Mock(**{'repeat.return_value': [(mocked_trs[0], 'cpu stats 0'), (mocked_trs[1], 'cpu stats 1')]})
    mocked_root = Mock(**{'findmeld.return_value': mocked_mid})
    # test call
    view.write_processor_statistics(mocked_root, [], [1, 2, 3])
    assert mocked_title.call_args_list == [call(mocked_trs[0], 1, 0), call(mocked_trs[1], 1, 1)]
    assert mocked_stats.call_args_list == [call(mocked_trs[0], 'cpu stats 0', [1, 2, 3]),
                                           call(mocked_trs[1], 'cpu stats 1', [1, 2, 3])]
    assert mocked_trs[0].attrib == {'class': 'brightened'}
    assert mocked_trs[1].attrib == {'class': 'shaded'}


def test_write_memory_statistics(mocker, view):
    """ Test the write_memory_statistics method. """
    mocked_common = mocker.patch.object(view, '_write_common_detailed_statistics')
    # replace root element
    mocked_root = Mock()
    # test method call
    view.write_memory_statistics(mocked_root, [1.523, 2.456], [1, 2, 3])
    assert mocked_common.call_args_list == [call(mocked_root, [1.523, 2.456], [1, 2, 3],
                                                 'mem_val_td_mid', 'mem_avg_td_mid',
                                                 'mem_slope_td_mid', 'mem_dev_td_mid')]


def test_write_network_io_single_title(view):
    """ Test the _write_network_io_single_title method. """
    # set context (meant to be set through render)
    view.view_ctx = Mock(**{'format_url.return_value': 'http://addr:port/index.html'})
    # replace root structure
    net_io_nic_a_mid = create_element()
    net_io_nic_td_mid = create_element({'net_io_nic_a_mid': net_io_nic_a_mid})
    tr_elt = create_element({'net_io_nic_td_mid': net_io_nic_td_mid})
    # in first call, elt is not the first line (rowspan False)
    view._write_network_io_single_title(tr_elt, 'eth0', 'lo', False, True)
    assert tr_elt.findmeld.call_args_list == [call('net_io_nic_td_mid')]
    assert net_io_nic_td_mid.attrib == {'class': ''}
    assert not net_io_nic_td_mid.findmeld.called
    assert net_io_nic_a_mid.attrib == {'class': ''}
    assert net_io_nic_td_mid.replace.call_args_list == [call('')]
    tr_elt.reset_all()
    # in second call, elt is the first line (rowspan True), shaded and is not the selected interface
    view._write_network_io_single_title(tr_elt, 'eth0', 'lo', True, True)
    assert tr_elt.findmeld.call_args_list == [call('net_io_nic_td_mid')]
    assert net_io_nic_td_mid.attrib == {'class': 'shaded', 'rowspan': '2'}
    assert net_io_nic_td_mid.findmeld.call_args_list == [call('net_io_nic_a_mid')]
    assert net_io_nic_a_mid.attrib == {'class': ''}
    assert net_io_nic_a_mid.attributes.call_args_list == [call(href='http://addr:port/index.html')]
    assert not net_io_nic_td_mid.replace.called
    # reset context
    tr_elt.reset_all()
    # in third call, elt is the first line (rowspan True), not shaded and is the selected interface
    view._write_network_io_single_title(tr_elt, 'lo', 'lo', True, False)
    assert tr_elt.findmeld.call_args_list == [call('net_io_nic_td_mid')]
    assert net_io_nic_td_mid.attrib == {'class': 'brightened', 'rowspan': '2'}
    assert net_io_nic_td_mid.findmeld.call_args_list == [call('net_io_nic_a_mid')]
    assert net_io_nic_a_mid.attrib == {'class': 'button off active'}
    assert not net_io_nic_a_mid.attributes.called
    assert not net_io_nic_td_mid.replace.called


def test_write_network_io_single_statistics(mocker, view):
    """ Test the _write_network_io_single_statistics method. """
    mocked_common = mocker.patch.object(view, '_write_common_detailed_statistics')
    # replace root structure
    net_io_rxtx_td_mid = create_element()
    tr_elt = create_element({'net_io_rxtx_td_mid': net_io_rxtx_td_mid})
    # test send bytes
    view._write_network_io_single_statistics(tr_elt, [1.523, 2.456], [1, 2, 3], False)
    assert tr_elt.findmeld.call_args_list == [call('net_io_rxtx_td_mid')]
    assert net_io_rxtx_td_mid.content.call_args_list == [call('Tx')]
    assert mocked_common.call_args_list == [call(tr_elt, [1.523, 2.456], [1, 2, 3],
                                                 'net_io_val_td_mid', 'net_io_avg_td_mid',
                                                 'net_io_slope_td_mid', 'net_io_dev_td_mid')]
    tr_elt.reset_all()
    mocker.resetall()
    # test recv bytes
    view._write_network_io_single_statistics(tr_elt, [1.523, 2.456], [1, 2, 3], True)
    assert tr_elt.findmeld.call_args_list == [call('net_io_rxtx_td_mid')]
    assert net_io_rxtx_td_mid.content.call_args_list == [call('Rx')]
    assert mocked_common.call_args_list == [call(tr_elt, [1.523, 2.456], [1, 2, 3],
                                                 'net_io_val_td_mid', 'net_io_avg_td_mid',
                                                 'net_io_slope_td_mid', 'net_io_dev_td_mid')]


def test_write_network_io_statistics(mocker, view):
    """ Test the write_network_io_statistics method. """
    mocked_stats = mocker.patch.object(view, '_write_network_io_single_statistics')
    mocked_title = mocker.patch.object(view, '_write_network_io_single_title')
    # set context (meant to be set through render)
    view.view_ctx = Mock(nic_name='eth0')
    # build root structure
    net_io_tr_elts = [create_element() for _ in range(4)]
    net_io_tr_mid = create_element()
    net_io_tr_mid.repeat.return_value = [(net_io_tr_elts[0], ('lo', [1, 2, 3], 'lo recv')),
                                         (net_io_tr_elts[1], ('lo', [1, 2, 3], 'lo sent')),
                                         (net_io_tr_elts[2], ('eth0', [2, 3], 'eth0 recv')),
                                         (net_io_tr_elts[3], ('eth0', [2, 3], 'eth0 sent'))]
    tr_elt = create_element({'net_io_tr_mid': net_io_tr_mid})
    # test method with dummy stats
    dummy_io_stats = {'lo': [[1, 2, 3], ['lo recv', 'lo sent']],
                      'eth0': [[2, 3], ['eth0 recv', 'eth0 sent']]}
    view.write_network_io_statistics(tr_elt, dummy_io_stats)
    # check calls
    assert tr_elt.findmeld.call_args_list == [call('net_io_tr_mid')]
    assert net_io_tr_mid.repeat.call_args_list == [call([('lo', [1, 2, 3], 'lo recv'),
                                                         ('lo', [1, 2, 3], 'lo sent'),
                                                         ('eth0', [2, 3], 'eth0 recv'),
                                                         ('eth0', [2, 3], 'eth0 sent')])]
    assert net_io_tr_elts[0].attrib['class'] == 'brightened'
    assert net_io_tr_elts[1].attrib['class'] == 'brightened'
    assert net_io_tr_elts[2].attrib['class'] == 'shaded'
    assert net_io_tr_elts[3].attrib['class'] == 'shaded'
    assert mocked_title.call_args_list == [call(net_io_tr_elts[0], 'eth0', 'lo', True, False),
                                           call(net_io_tr_elts[1], 'eth0', 'lo', False, False),
                                           call(net_io_tr_elts[2], 'eth0', 'eth0', True, True),
                                           call(net_io_tr_elts[3], 'eth0', 'eth0', False, True)]
    assert mocked_stats.call_args_list == [call(net_io_tr_elts[0], 'lo recv', [1, 2, 3], True),
                                           call(net_io_tr_elts[1], 'lo sent', [1, 2, 3], False),
                                           call(net_io_tr_elts[2], 'eth0 recv', [2, 3], True),
                                           call(net_io_tr_elts[3], 'eth0 sent', [2, 3], False)]


def test_write_disk_io_single_title(view):
    """ Test the _write_disk_io_single_title method. """
    # set context (meant to be set through render)
    view.view_ctx = Mock(**{'format_url.return_value': 'http://addr:port/index.html'})
    # replace root structure
    disk_io_device_a_mid = create_element()
    disk_io_device_td_mid = create_element({'disk_io_device_a_mid': disk_io_device_a_mid})
    tr_elt = create_element({'disk_io_device_td_mid': disk_io_device_td_mid})
    # in first call, elt is not the first line (rowspan False)
    view._write_disk_io_single_title(tr_elt, 'sdb', 'sda', False, True)
    assert tr_elt.findmeld.call_args_list == [call('disk_io_device_td_mid')]
    assert disk_io_device_td_mid.attrib == {'class': ''}
    assert not disk_io_device_td_mid.findmeld.called
    assert disk_io_device_a_mid.attrib == {'class': ''}
    assert disk_io_device_td_mid.replace.call_args_list == [call('')]
    tr_elt.reset_all()
    # in second call, elt is the first line (rowspan True), shaded and is not the selected interface
    view._write_disk_io_single_title(tr_elt, 'sdb', 'sda', True, True)
    assert tr_elt.findmeld.call_args_list == [call('disk_io_device_td_mid')]
    assert disk_io_device_td_mid.attrib == {'class': 'shaded', 'rowspan': '2'}
    assert disk_io_device_td_mid.findmeld.call_args_list == [call('disk_io_device_a_mid')]
    assert disk_io_device_a_mid.attrib == {'class': ''}
    assert disk_io_device_a_mid.attributes.call_args_list == [call(href='http://addr:port/index.html')]
    assert not disk_io_device_td_mid.replace.called
    tr_elt.reset_all()
    # in third call, elt is the first line (rowspan True), not shaded and is the selected interface
    view._write_disk_io_single_title(tr_elt, 'sda', 'sda', True, False)
    assert tr_elt.findmeld.call_args_list == [call('disk_io_device_td_mid')]
    assert disk_io_device_td_mid.attrib == {'class': 'brightened', 'rowspan': '2'}
    assert disk_io_device_td_mid.findmeld.call_args_list == [call('disk_io_device_a_mid')]
    assert disk_io_device_a_mid.attrib == {'class': 'button off active'}
    assert not disk_io_device_a_mid.attributes.called
    assert not disk_io_device_td_mid.replace.called


def test_write_disk_io_single_statistics(mocker, view):
    """ Test the _write_disk_io_single_statistics method. """
    mocked_common = mocker.patch.object(view, '_write_common_detailed_statistics')
    # replace root structure
    disk_io_rw_td_mid = create_element()
    tr_elt = create_element({'disk_io_rw_td_mid': disk_io_rw_td_mid})
    # test send bytes
    view._write_disk_io_single_statistics(tr_elt, [1.523, 2.456], [1, 2, 3], False)
    assert tr_elt.findmeld.call_args_list == [call('disk_io_rw_td_mid')]
    assert disk_io_rw_td_mid.content.call_args_list == [call('W')]
    assert mocked_common.call_args_list == [call(tr_elt, [1.523, 2.456], [1, 2, 3],
                                                 'disk_io_val_td_mid', 'disk_io_avg_td_mid',
                                                 'disk_io_slope_td_mid', 'disk_io_dev_td_mid')]
    tr_elt.reset_all()
    mocker.resetall()
    # test recv bytes
    view._write_disk_io_single_statistics(tr_elt, [1.523, 2.456], [1, 2, 3], True)
    assert tr_elt.findmeld.call_args_list == [call('disk_io_rw_td_mid')]
    assert disk_io_rw_td_mid.content.call_args_list == [call('R')]
    assert mocked_common.call_args_list == [call(tr_elt, [1.523, 2.456], [1, 2, 3],
                                                 'disk_io_val_td_mid', 'disk_io_avg_td_mid',
                                                 'disk_io_slope_td_mid', 'disk_io_dev_td_mid')]


def test_write_disk_io_statistics(mocker, view):
    """ Test the write_disk_io_statistics method. """
    mocked_stats = mocker.patch.object(view, '_write_disk_io_single_statistics')
    mocked_title = mocker.patch.object(view, '_write_disk_io_single_title')
    # set context (meant to be set through render)
    view.view_ctx = Mock(device='sdb')
    # build root structure
    disk_io_tr_elts = [create_element() for _ in range(4)]
    disk_io_tr_mid = create_element()
    disk_io_tr_mid.repeat.return_value = [(disk_io_tr_elts[0], ('sda', [1, 2, 3], 'sda read')),
                                          (disk_io_tr_elts[1], ('sda', [1, 2, 3], 'sda write')),
                                          (disk_io_tr_elts[2], ('sdb', [2, 3], 'sdb read')),
                                          (disk_io_tr_elts[3], ('sdb', [2, 3], 'sdb write'))]
    tr_elt = create_element({'disk_io_tr_mid': disk_io_tr_mid})
    # test method with dummy stats
    dummy_io_stats = {'sda': [[1, 2, 3], ['sda read', 'sda write']],
                      'sdb': [[2, 3], ['sdb read', 'sdb write']]}
    view.write_disk_io_statistics(tr_elt, dummy_io_stats)
    # check calls
    assert tr_elt.findmeld.call_args_list == [call('disk_io_tr_mid')]
    assert disk_io_tr_mid.repeat.call_args_list == [call([('sda', [1, 2, 3], 'sda read'),
                                                          ('sda', [1, 2, 3], 'sda write'),
                                                          ('sdb', [2, 3], 'sdb read'),
                                                          ('sdb', [2, 3], 'sdb write')])]
    assert disk_io_tr_elts[0].attrib['class'] == 'brightened'
    assert disk_io_tr_elts[1].attrib['class'] == 'brightened'
    assert disk_io_tr_elts[2].attrib['class'] == 'shaded'
    assert disk_io_tr_elts[3].attrib['class'] == 'shaded'
    assert mocked_title.call_args_list == [call(disk_io_tr_elts[0], 'sdb', 'sda', True, False),
                                           call(disk_io_tr_elts[1], 'sdb', 'sda', False, False),
                                           call(disk_io_tr_elts[2], 'sdb', 'sdb', True, True),
                                           call(disk_io_tr_elts[3], 'sdb', 'sdb', False, True)]
    assert mocked_stats.call_args_list == [call(disk_io_tr_elts[0], 'sda read', [1, 2, 3], True),
                                           call(disk_io_tr_elts[1], 'sda write', [1, 2, 3], False),
                                           call(disk_io_tr_elts[2], 'sdb read', [2, 3], True),
                                           call(disk_io_tr_elts[3], 'sdb write', [2, 3], False)]


def test_write_disk_usage_single_title(view):
    """ Test the _write_disk_usage_single_title method. """
    # set context (meant to be set through render)
    view.view_ctx = Mock(**{'format_url.return_value': 'http://addr:port/index.html'})
    # replace root structure
    disk_usage_partition_a_mid = create_element()
    tr_elt = create_element({'disk_usage_partition_a_mid': disk_usage_partition_a_mid})
    # in first call, elt is not the selected partition
    view._write_disk_usage_single_title(tr_elt, '/root', '/')
    assert tr_elt.findmeld.call_args_list == [call('disk_usage_partition_a_mid')]
    assert disk_usage_partition_a_mid.attrib == {'class': ''}
    assert disk_usage_partition_a_mid.attributes.call_args_list == [call(href='http://addr:port/index.html')]
    tr_elt.reset_all()
    # in second call, elt is the selected partition
    view._write_disk_usage_single_title(tr_elt, '/root', '/root')
    assert tr_elt.findmeld.call_args_list == [call('disk_usage_partition_a_mid')]
    assert disk_usage_partition_a_mid.attrib == {'class': 'button off active'}
    assert not disk_usage_partition_a_mid.attributes.called


def test_write_disk_usage_single_statistics(mocker, view):
    """ Test the _write_disk_usage_single_statistics method. """
    mocked_common = mocker.patch.object(view, '_write_common_detailed_statistics')
    # replace root structure
    tr_elt = create_element()
    # test call
    view._write_disk_usage_single_statistics(tr_elt, [1.523, 2.456], [1, 2, 3])
    assert mocked_common.call_args_list == [call(tr_elt, [1.523, 2.456], [1, 2, 3],
                                                 'disk_usage_val_td_mid', 'disk_usage_avg_td_mid',
                                                 'disk_usage_slope_td_mid', 'disk_usage_dev_td_mid')]


def test_write_disk_usage_statistics(mocker, view):
    """ Test the write_disk_usage_statistics method. """
    mocked_stats = mocker.patch.object(view, '_write_disk_usage_single_statistics')
    mocked_title = mocker.patch.object(view, '_write_disk_usage_single_title')
    # set context (meant to be set through render)
    view.view_ctx = Mock(partition='/root')
    # build root structure
    disk_usage_tr_elts = [create_element() for _ in range(2)]
    disk_usage_tr_mid = create_element()
    disk_usage_tr_mid.repeat.return_value = [(disk_usage_tr_elts[0], ['/', [[1, 2, 3], ['disk usage stats /']]]),
                                             (disk_usage_tr_elts[1], ['/root', [[2, 3], ['disk usage stats /root']]])]
    mocked_root = create_element({'disk_usage_tr_mid': disk_usage_tr_mid})
    # test call
    view.write_disk_usage_statistics(mocked_root, Mock())
    assert mocked_title.call_args_list == [call(disk_usage_tr_elts[0], '/root', '/'),
                                           call(disk_usage_tr_elts[1], '/root', '/root')]
    assert mocked_stats.call_args_list == [call(disk_usage_tr_elts[0], 'disk usage stats /', [1, 2, 3]),
                                           call(disk_usage_tr_elts[1], 'disk usage stats /root', [2, 3])]
    assert disk_usage_tr_elts[0].attrib == {'class': 'brightened'}
    assert disk_usage_tr_elts[1].attrib == {'class': 'shaded'}


def test_write_disk_statistics(mocker, view):
    """ Test the write_disk_statistics method. """
    mocked_usage = mocker.patch.object(view, 'write_disk_usage_statistics')
    mocked_io = mocker.patch.object(view, 'write_disk_io_statistics')
    # set context (meant to be set through render)
    stats = Mock(disk_usage='disk usage', disk_io='disk io')
    # create xhtml structure
    disk_io_div_mid = create_element()
    disk_io_view_mid = create_element()
    disk_usage_div_mid = create_element()
    disk_usage_view_mid = create_element()
    contents_elt = create_element({'disk_io_div_mid': disk_io_div_mid, 'disk_usage_div_mid': disk_usage_div_mid,
                                   'disk_io_view_mid': disk_io_view_mid, 'disk_usage_view_mid': disk_usage_view_mid})
    # Disk IO selection
    view.view_ctx = Mock(disk_stats='io', **{'format_url.return_value': 'an url'})
    view.write_disk_statistics(contents_elt, stats)
    assert mocked_io.call_args_list == [call(contents_elt, 'disk io')]
    assert not mocked_usage.called
    assert contents_elt.findmeld.call_args_list == [call('disk_usage_div_mid'), call('disk_io_view_mid')]
    assert disk_usage_div_mid.replace.call_args_list == [call('')]
    assert disk_io_view_mid.attributes.call_args_list == [call(href='an url')]
    assert not disk_io_div_mid.replace.called
    assert not disk_usage_view_mid.attributes.called
    mocker.resetall()
    contents_elt.reset_all()
    # Disk IO selection
    view.view_ctx.disk_stats = 'usage'
    view.write_disk_statistics(contents_elt, stats)
    assert mocked_usage.call_args_list == [call(contents_elt, 'disk usage')]
    assert not mocked_io.called
    assert contents_elt.findmeld.call_args_list == [call('disk_io_div_mid'), call('disk_usage_view_mid')]
    assert disk_io_div_mid.replace.call_args_list == [call('')]
    assert disk_usage_view_mid.attributes.call_args_list == [call(href='an url')]
    assert not disk_usage_div_mid.replace.called
    assert not disk_io_view_mid.attributes.called


def test_write_cpu_image(mocker, view):
    """ Test the _write_cpu_image method. """
    mocked_export = mocker.patch('supvisors.plot.StatisticsPlot.export_image')
    mocked_plot = mocker.patch('supvisors.plot.StatisticsPlot.add_plot')
    mocked_time = mocker.patch('supvisors.plot.StatisticsPlot.add_timeline')
    # set context (meant to be set through render)
    view.view_ctx = Mock(cpu_id=0, **{'cpu_id_to_string.return_value': ViewContext.cpu_id_to_string(0)})
    # just test calls to StatisticsPlot
    dummy_cpu_stats = ['#all stats', '#0 stats', '#1 stats']
    dummy_times_stats = [1, 2, 3]
    view._write_cpu_image(dummy_cpu_stats, dummy_times_stats)
    assert mocked_time.call_args_list == [call(dummy_times_stats)]
    assert mocked_plot.call_args_list == [call('CPU #all', '%', '#all stats')]
    assert mocked_export.call_args_list == [call(host_cpu_img)]


def test_write_mem_image(mocker, view):
    """ Test the _write_mem_image method. """
    mocked_export = mocker.patch('supvisors.plot.StatisticsPlot.export_image')
    mocked_plot = mocker.patch('supvisors.plot.StatisticsPlot.add_plot')
    mocked_time = mocker.patch('supvisors.plot.StatisticsPlot.add_timeline')
    # just test calls to StatisticsPlot
    dummy_mem_stats = ['mem 1', 'mem 2']
    dummy_times_stats = [1, 2, 3]
    view._write_mem_image(dummy_mem_stats, dummy_times_stats)
    assert mocked_time.call_args_list == [call(dummy_times_stats)]
    assert mocked_plot.call_args_list == [call('MEM', '%', dummy_mem_stats)]
    assert mocked_export.call_args_list == [call(host_mem_img)]


def test_write_net_io_image(mocker, view):
    """ Test the _write_net_io_image method. """
    mocked_export = mocker.patch('supvisors.plot.StatisticsPlot.export_image')
    mocked_plot = mocker.patch('supvisors.plot.StatisticsPlot.add_plot')
    mocked_time = mocker.patch('supvisors.plot.StatisticsPlot.add_timeline')
    # set context (meant to be set through render)
    view.view_ctx = Mock(nic_name='eth0')
    # just test calls to StatisticsPlot
    dummy_io_stats = {'lo': [[1, 2, 3, 4], ['lo recv', 'lo sent']],
                      'eth0': [[2, 3, 4], ['eth0 recv', 'eth0 sent']]}
    view._write_net_io_image(dummy_io_stats)
    assert mocked_time.call_args_list == [call([2, 3, 4])]
    assert mocked_plot.call_args_list == [call('eth0 received', 'kbits/s', 'eth0 recv'),
                                          call('eth0 sent', 'kbits/s', 'eth0 sent')]
    assert mocked_export.call_args_list == [call(host_net_io_img)]


def test_write_disk_image(mocker, view):
    """ Test the _write_disk_image method. """
    mocked_usage = mocker.patch.object(view, '_write_disk_usage_image')
    mocked_io = mocker.patch.object(view, '_write_disk_io_image')
    # set context (meant to be set through render)
    stats = Mock(disk_usage='disk usage', disk_io='disk io')
    # Disk IO selection
    view.view_ctx = Mock(disk_stats='io')
    view._write_disk_image(stats)
    assert mocked_io.call_args_list == [call('disk io')]
    assert not mocked_usage.called
    mocker.resetall()
    # Disk IO selection
    view.view_ctx.disk_stats = 'usage'
    view._write_disk_image(stats)
    assert mocked_usage.call_args_list == [call('disk usage')]
    assert not mocked_io.called


def test_write_disk_io_image(mocker, view):
    """ Test the _write_disk_io_image method. """
    mocked_export = mocker.patch('supvisors.plot.StatisticsPlot.export_image')
    mocked_plot = mocker.patch('supvisors.plot.StatisticsPlot.add_plot')
    mocked_time = mocker.patch('supvisors.plot.StatisticsPlot.add_timeline')
    # set context (meant to be set through render)
    view.view_ctx = Mock(device='sda')
    # just test calls to StatisticsPlot
    dummy_io_stats = {'sda': [[1, 2, 3, 4], ['sda read', 'sda written']],
                      'sdb': [[2, 3, 4], ['sdb read', 'sdb written']]}
    view._write_disk_io_image(dummy_io_stats)
    assert mocked_time.call_args_list == [call([1, 2, 3, 4])]
    assert mocked_plot.call_args_list == [call('sda read', 'kbits/s', 'sda read'),
                                          call('sda written', 'kbits/s', 'sda written')]
    assert mocked_export.call_args_list == [call(host_disk_io_img)]


def test_write_disk_usage_image(mocker, view):
    """ Test the _write_disk_usage_image method. """
    mocked_export = mocker.patch('supvisors.plot.StatisticsPlot.export_image')
    mocked_plot = mocker.patch('supvisors.plot.StatisticsPlot.add_plot')
    mocked_time = mocker.patch('supvisors.plot.StatisticsPlot.add_timeline')
    # set context (meant to be set through render)
    view.view_ctx = Mock(partition='/root')
    # just test calls to StatisticsPlot
    dummy_io_stats = {'/': [[1, 2, 3, 4], ['/ percent']],
                      '/root': [[2, 3, 4], ['/root percent']]}
    view._write_disk_usage_image(dummy_io_stats)
    assert mocked_time.call_args_list == [call([2, 3, 4])]
    assert mocked_plot.call_args_list == [call('/root', '%', '/root percent')]
    assert mocked_export.call_args_list == [call(host_disk_usage_img)]



# ======================================================================
# Copyright 2016 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import os
import random
import socket
from socket import getfqdn, gethostname
from unittest.mock import Mock

from supervisor.loggers import getLogger, handle_stdout, Logger
from supervisor.rpcinterface import SupervisorNamespaceRPCInterface
from supervisor.states import STOPPED_STATES, SupervisorStates, ProcessStates

import supvisors
from supvisors.context import Context
from supvisors.initializer import LOGGER_FORMAT
from supvisors.internal_com.mapper import SupvisorsMapper
from supvisors.options import SupvisorsOptions, SupvisorsServerOptions
from supvisors.rpcinterface import RPCInterface
from supvisors.statscollector import StatisticsCollectorProcess
from supvisors.statscompiler import HostStatisticsCompiler, ProcStatisticsCompiler
from supvisors.supervisordata import SupervisorData
from supvisors.supervisorupdater import SupervisorUpdater
from supvisors.utils import extract_process_info


class MockedSupvisors:
    """ Simple supvisors with all dummies. """

    def __init__(self, supervisord, config):
        """ Use mocks when not possible to use real structures. """
        self.logger = Mock(spec=Logger, level=10, handlers=[Mock(level=10)])
        self.options = SupvisorsOptions(supervisord, self.logger, **config)
        self.options.rules_files = [config['rules_files']]
        # mock the supervisord source
        self.supervisor_data = SupervisorData(self, supervisord)
        self.supervisor_updater = SupervisorUpdater(self)
        self.mapper = SupvisorsMapper(self)
        host_name = gethostname()
        fqdn = getfqdn()
        identifiers = ['10.0.0.1', '10.0.0.2', '10.0.0.3', '10.0.0.4', '10.0.0.5',
                       f'<{host_name}>{fqdn}:25000', f'<test>{fqdn}:15000']
        self.mapper.configure(identifiers, {'supvisors_test'}, [])
        self.server_options = SupvisorsServerOptions(self)
        # set real statistics collector and compilers
        self.stats_collector = StatisticsCollectorProcess(self)
        self.host_compiler = HostStatisticsCompiler(self)
        self.process_compiler = ProcStatisticsCompiler(self.options, self.logger)
        # build context from node mapper
        self.context = Context(self)
        # mock by spec
        from supvisors.commander import Starter, Stopper, StarterModel
        from supvisors.strategy import RunningFailureHandler
        from supvisors.statemachine import FiniteStateMachine
        from supvisors.listener import SupervisorListener
        from supvisors.sparser import Parser
        from supvisors.internal_com.rpchandler import RpcHandler
        self.starter = Mock(spec=Starter)
        self.starter_model = Mock(spec=StarterModel)
        self.stopper = Mock(spec=Stopper)
        self.failure_handler = Mock(spec=RunningFailureHandler)
        self.fsm = Mock(spec=FiniteStateMachine, redeploy_mark=False)
        self.listener = Mock(spec=SupervisorListener)
        self.parser = Mock(spec=Parser)
        # should be set in listener
        self.rpc_handler = Mock(spec=RpcHandler)
        self.rpc_handler.__init__()
        self.discovery_handler = None
        self.external_publisher = None


class DummyRpcHandler:
    """ Simple supervisord RPC handler with dummy attributes. """

    def __init__(self):
        self.handler = Mock(rpcinterface=Mock(system=Mock(rpc_name='system_RPC'),
                                              supervisor=Mock(rpc_name='supervisor_RPC'),
                                              supvisors=Mock(rpc_name='supvisors_RPC')))


class DummyRpcInterface:
    """ Simple RPC mock. """

    def __init__(self, supvisors):
        # create rpc interfaces to have a skeleton
        self.supervisor = SupervisorNamespaceRPCInterface(DummySupervisor())
        self.supvisors = RPCInterface(supvisors)


class DummyHttpServer:
    """ Simple supervisord RPC handler with dummy attributes. """

    def __init__(self):
        self.socket = Mock()
        self.handlers = [DummyRpcHandler(),
                         Mock(handler_name='tail_handler'),
                         Mock(handler_name='main_tail_handler'),
                         Mock(handler_name='ui_handler'),
                         Mock(handler_name='default_handler')]


class DummyServerOptions:
    """ Simple supervisord server options with dummy attributes. """

    def __init__(self):
        # build a fake server config
        server_config = {'section': 'inet_http_server',
                         'family': socket.AF_INET,
                         'host': gethostname(),
                         'port': 25000,
                         'username': 'user',
                         'password': 'p@$$w0rd'}
        self.configfile = 'supervisord.conf'
        self.httpserver = DummyHttpServer()
        self.server_configs = [server_config]
        self.here = '.'
        self.environ_expansions = {}
        self.identifier = gethostname()
        self.serverurl = 'unix://tmp/supervisor.sock'
        self.mood = SupervisorStates.RUNNING
        self.nodaemon = True
        self.silent = False
        # add silent logger to test create_logger
        self.logfile = 'supervisor.log'
        self.logger = getLogger()
        handle_stdout(self.logger, LOGGER_FORMAT)
        self.logger.log = Mock()
        # build a fake http config
        self.httpservers = [(server_config, self.httpserver)]
        # prepare storage for close_httpservers test
        self.storage = None

    def close_httpservers(self):
        self.storage = self.httpservers


class DummyProcessConfig:
    """ Simple supervisor process config with simple attributes. """

    def __init__(self, name: str, command: str, autorestart: bool, stdout_logfile: bool, stderr_logfile: bool):
        self.name = name
        self.command = command
        self.autorestart = autorestart
        self.startsecs = 5
        self.stopwaitsecs = 10
        self.stdout_logfile = stdout_logfile
        self.stderr_logfile = stderr_logfile


class DummyProcess:
    """ Simple supervisor process with simple attributes. """

    def __init__(self, name: str, command: str, autorestart: bool, stdout_logfile: bool, stderr_logfile: bool):
        self.state = ProcessStates.STOPPED
        self.spawnerr = ''
        self.laststart = 1234
        self.config = DummyProcessConfig(name, command, autorestart, stdout_logfile, stderr_logfile)

    def give_up(self):
        self.state = ProcessStates.FATAL

    def transition(self):
        self.state = ProcessStates.STARTING


class DummySupervisor:
    """ Simple supervisor with simple attributes. """

    def __init__(self):
        self.options = DummyServerOptions()
        dummy_process_1 = DummyProcess('dummy_process_1', 'ls', True, True, False)
        dummy_process_2 = DummyProcess('dummy_process_2', 'cat', False, False, True)
        self.process_groups = {'dummy_application': Mock(config=Mock(name='dummy_application'),
                                                         processes={'dummy_process_1': dummy_process_1,
                                                                    'dummy_process_2': dummy_process_2})}


class DummyHttpContext:
    """ Simple HTTP context for web ui views. """

    def __init__(self, template=None):
        supvisors_path = next(iter(supvisors.__path__), '.')
        self.template = os.path.join(supvisors_path, template) if template else None
        self.supervisord = DummySupervisor()
        # create form and response
        self.form = {'SERVER_URL': 'http://10.0.0.1:7777',
                     'SERVER_PORT': 7777,
                     'PATH_TRANSLATED': '/index.html',
                     'action': 'test',
                     'ident': '10.0.0.4:25000',
                     'message': 'hi chaps',
                     'gravity': 'none',
                     'namespec': 'dummy_proc',
                     'processname': 'dummy_proc',
                     'appname': 'dummy_appli',
                     'nic': 'eth0',
                     'period': 5.1,
                     'auto': 'false',
                     'diskstats': 'io',
                     'partition': '/', 'device': 'sda'}
        self.response = {'headers': {'Location': None}}


# note that all dates ('now') are different
ProcessInfoDatabase = [
    {'description': '', 'pid': 80886, 'stderr_logfile': '', 'stop': 1473888084,
     'logfile': './log/late_segv_cliche01.log', 'exitstatus': 0, 'spawnerr': '', 'now': 1473888091,
     'group': 'crash', 'name': 'late_segv', 'statename': 'STARTING', 'start': 1473888089, 'state': 10,
     'stdout_logfile': './log/late_segv_cliche01.log'},
    {'description': 'Exited too quickly (process log may have details)', 'pid': 0, 'stderr_logfile': '',
     'stop': 1473888156, 'logfile': './log/segv_cliche01.log', 'exitstatus': 0,
     'spawnerr': 'Exited too quickly (process log may have details)', 'now': 1473888156,
     'group': 'crash', 'name': 'segv', 'statename': 'BACKOFF', 'start': 1473888155, 'state': 30,
     'stdout_logfile': './log/segv_cliche01.log'},
    {'description': 'Sep 14 05:18 PM', 'pid': 0, 'stderr_logfile': '', 'stop': 1473888937,
     'logfile': './log/firefox_cliche01.log', 'exitstatus': 0, 'spawnerr': '', 'now': 1473888161,
     'group': 'firefox', 'name': 'firefox', 'statename': 'EXITED', 'start': 1473887932, 'state': 100,
     'stdout_logfile': './log/firefox_cliche01.log'},
    {'description': 'pid 80877, uptime 0:01:20', 'pid': 80877, 'stderr_logfile': '', 'stop': 0,
     'logfile': './log/xclock_cliche01.log', 'exitstatus': 0, 'spawnerr': '', 'now': 1473888166,
     'group': 'sample_test_1', 'name': 'xclock', 'statename': 'STOPPING', 'start': 1473888078, 'state': 40,
     'stdout_logfile': './log/xclock_cliche01.log'},
    {'description': 'pid 80879, uptime 0:01:19', 'pid': 80879, 'stderr_logfile': '', 'stop': 0,
     'logfile': './log/xfontsel_cliche01.log', 'exitstatus': 0, 'spawnerr': '', 'now': 1473888171,
     'group': 'sample_test_1', 'name': 'xfontsel', 'statename': 'RUNNING', 'start': 1473888079, 'state': 20,
     'stdout_logfile': './log/xfontsel_cliche01.log'},
    {'description': 'Sep 14 05:21 PM', 'pid': 0, 'stderr_logfile': '', 'stop': 1473888104,
     'logfile': './log/xlogo_cliche01.log', 'exitstatus': -1, 'spawnerr': '', 'now': 1473888176,
     'group': 'sample_test_1', 'name': 'xlogo', 'statename': 'STOPPED', 'start': 1473888085, 'state': 0,
     'stdout_logfile': './log/xlogo_cliche01.log'},
    {'description': 'No resource available', 'pid': 0, 'stderr_logfile': '', 'stop': 0,
     'logfile': './log/sleep_cliche01.log', 'exitstatus': 0, 'spawnerr': 'No resource available',
     'now': 1473888181, 'group': 'sample_test_2', 'name': 'sleep', 'statename': 'FATAL', 'start': 0, 'state': 200,
     'stdout_logfile': './log/sleep_cliche01.log'},
    {'description': 'Sep 14 05:22 PM', 'pid': 0, 'stderr_logfile': '', 'stop': 1473888130,
     'logfile': './log/xeyes_cliche01.log', 'exitstatus': 0, 'spawnerr': '', 'now': 1473888186,
     'group': 'sample_test_2', 'name': 'yeux_00', 'statename': 'EXITED', 'start': 1473888086, 'state': 100,
     'stdout_logfile': './log/xeyes_cliche01.log'},
    {'description': 'pid 80882, uptime 0:01:12', 'pid': 80882, 'stderr_logfile': '', 'stop': 0,
     'logfile': './log/xeyes_cliche01.log', 'exitstatus': 0, 'spawnerr': '', 'now': 1473888196,
     'group': 'sample_test_2', 'name': 'yeux_01', 'statename': 'RUNNING', 'start': 1473888086, 'state': 20,
     'stdout_logfile': './log/xeyes_cliche01.log'}]


def extract_and_complete(info):
    """ Provide payload as processed by Supvisors. """
    extracted_info = extract_process_info(info)
    extracted_info.update({'startsecs': 0, 'stopwaitsecs': 0, 'extra_args': '', 'disabled': False})
    extracted_info.update({'now_monotonic': info['now'] - 1e6,
                           'start_monotonic': info['start'] - 1e6,
                           'stop_monotonic': info['stop'] - 1e6})
    if info['name'].startswith('yeux'):
        program_name, process_index = info['name'].split('_')
        extracted_info.update({'program_name': program_name, 'process_index': int(process_index)})
    else:
        extracted_info.update({'program_name': info['name'], 'process_index': 0})
    return extracted_info


def database_copy():
    """ Return a copy of the whole database. """
    return [extract_and_complete(info) for info in ProcessInfoDatabase]


def any_process_info():
    """ Return a copy of any process in database. """
    info = random.choice(ProcessInfoDatabase)
    return extract_and_complete(info)


def any_stopped_process_info():
    """ Return a copy of any stopped process in database. """
    info = random.choice([info for info in ProcessInfoDatabase if info['state'] in STOPPED_STATES])
    return extract_and_complete(info)


def any_process_info_by_state(state):
    """ Return a copy of any process in state 'state' in database. """
    info = random.choice([info for info in ProcessInfoDatabase if info['state'] == state])
    return extract_and_complete(info)


def process_info_by_name(name):
    """ Return a copy of a process named 'name' in database. """
    return next((extract_and_complete(info) for info in ProcessInfoDatabase if info['name'] == name), None)



# ======================================================================
# Copyright 2020 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

from random import shuffle
from unittest.mock import call, Mock

import pytest
from supervisor.web import MeldView, StatusView

from supvisors.ttypes import ApplicationStates
from supvisors.web.viewhandler import ViewHandler
from supvisors.web.viewprocinstance import *
from supvisors.web.webutils import PROC_INSTANCE_PAGE
from .base import DummyHttpContext, ProcessInfoDatabase, process_info_by_name
from .conftest import create_application, create_process, create_element


@pytest.fixture
def http_context(supvisors):
    """ Fixture for a consistent mocked HTTP context provided by Supervisor. """
    http_context = DummyHttpContext('ui/proc_instance.html')
    http_context.supervisord.supvisors = supvisors
    supvisors.supervisor_data.supervisord = http_context.supervisord
    return http_context


@pytest.fixture
def view(http_context):
    """ Return the instance to test. """
    # create the instance to be tested
    return ProcInstanceView(http_context)


def test_init(view):
    """ Test the values set at construction of ProcInstanceView. """
    # test instance inheritance
    for klass in [SupvisorsInstanceView, StatusView, ViewHandler, MeldView]:
        assert isinstance(view, klass)
    # test default page name
    assert view.page_name == PROC_INSTANCE_PAGE


def test_handle_parameters(mocker, view):
    """ Test the handle_parameters method. """
    mocked_handle = mocker.patch('supvisors.web.viewhandler.ViewHandler.handle_parameters')
    view.view_ctx = Mock()
    view.handle_parameters()
    assert mocked_handle.call_args_list == [call()]
    assert view.view_ctx.set_message_default.call_args_list == [call('The Supvisors rules do NOT apply here', Warn)]


def test_write_options(mocker, view):
    """ Test the SupvisorsInstanceView.write_options method. """
    mocked_period = mocker.patch.object(view, 'write_periods')
    mocked_switch = mocker.patch.object(view, 'write_view_switch')
    option_card_mid = create_element()
    option_line_mid = create_element()
    period_div_mid = create_element()
    view_div_mid = create_element()
    mocked_header = create_element({'option_card_mid': option_card_mid, 'option_line_mid': option_line_mid,
                                    'period_div_mid': period_div_mid, 'view_div_mid': view_div_mid})
    # test all statistics enabled
    assert view.has_host_statistics
    assert view.has_process_statistics
    view.write_options(mocked_header)
    assert mocked_period.call_args_list == [call(mocked_header)]
    assert mocked_switch.call_args_list == [call(mocked_header)]
    assert not mocked_header.findmeld.called
    assert not option_card_mid.replace.called
    assert not option_line_mid.replace.called
    assert not period_div_mid.replace.called
    assert not view_div_mid.replace.called
    mocker.resetall()
    mocked_header.reset_all()
    # test process statistics only
    view.has_process_statistics = True
    view.has_host_statistics = False
    view.write_options(mocked_header)
    assert mocked_period.call_args_list == [call(mocked_header)]
    assert not mocked_switch.called
    assert mocked_header.findmeld.call_args_list == [call('view_div_mid')]
    assert not option_card_mid.replace.called
    assert not option_line_mid.replace.called
    assert not period_div_mid.replace.called
    assert view_div_mid.replace.call_args_list == [call('')]
    mocker.resetall()
    mocked_header.reset_all()
    # test host statistics only
    view.has_process_statistics = False
    view.has_host_statistics = True
    view.write_options(mocked_header)
    assert not mocked_period.called
    assert mocked_switch.call_args_list == [call(mocked_header)]
    assert mocked_header.findmeld.call_args_list == [call('period_div_mid')]
    assert not option_card_mid.replace.called
    assert not option_line_mid.replace.called
    assert period_div_mid.replace.call_args_list == [call('')]
    assert not view_div_mid.replace.called
    mocker.resetall()
    mocked_header.reset_all()
    # test no statistics
    view.has_process_statistics = False
    view.has_host_statistics = False
    view.write_options(mocked_header)
    assert not mocked_period.called
    assert not mocked_switch.called
    assert mocked_header.findmeld.call_args_list == [call('option_card_mid'), call('option_line_mid')]
    assert option_card_mid.replace.call_args_list == [call('')]
    assert option_line_mid.replace.call_args_list == [call('')]
    assert not period_div_mid.replace.called
    assert not view_div_mid.replace.called


def test_write_view_switch(supvisors, view):
    """ Test the SupvisorsInstanceView.write_view_switch method. """
    # set context (meant to be set through constructor and render)
    view.view_ctx = Mock(**{'format_url.return_value': 'an url'})
    supvisors.mapper.local_identifier = '10.0.0.1:25000'
    # build root structure
    mocked_process_view_mid = create_element()
    mocked_host_view_mid = create_element()
    mocked_header = create_element({'process_view_a_mid': mocked_process_view_mid,
                                    'host_view_a_mid': mocked_host_view_mid})
    # test call
    view.write_view_switch(mocked_header)
    assert mocked_header.findmeld.call_args_list == [call('host_view_a_mid')]
    assert view.view_ctx.format_url.call_args_list == [call('', HOST_INSTANCE_PAGE)]
    assert not mocked_process_view_mid.attributes.called
    assert mocked_host_view_mid.attributes.call_args_list == [call(href='an url')]
    assert mocked_host_view_mid.content.call_args_list == [call('10.0.0.1')]


def test_write_contents(mocker, view):
    """ Test the ProcInstanceView.write_contents method. """
    mocked_stats = mocker.patch.object(view, 'write_process_statistics')
    mocked_table = mocker.patch.object(view, 'write_process_table')
    mocked_data = mocker.patch.object(view, 'get_process_data',
                                      side_effect=(([{'namespec': 'dummy'}], []),
                                                   ([{'namespec': 'dummy'}], [{'namespec': 'dummy_proc'}]),
                                                   ([{'namespec': 'dummy'}], [{'namespec': 'dummy_proc'}]),
                                                   ([{'namespec': 'dummy'}], [{'namespec': 'dummy_proc'}]),
                                                   ([{'namespec': 'dummy_proc'}], [{'namespec': 'dummy'}])))
    # patch context
    view.view_ctx = Mock(process_name=None, local_identifier='10.0.0.1',
                         **{'get_process_status.return_value': None})
    # patch the meld elements
    contents_elt = create_element()
    # test call with no process selected
    view.write_contents(contents_elt)
    assert mocked_data.call_args_list == [call()]
    assert mocked_table.call_args_list == [call(contents_elt, [{'namespec': 'dummy'}], [])]
    assert mocked_stats.call_args_list == [call(contents_elt, {})]
    mocker.resetall()
    # test call with process selected and no corresponding status
    # process set in excluded_list but not passed to write_process_statistics because unselected due to missing status
    view.view_ctx.process_name = 'dummy_proc'
    view.write_contents(contents_elt)
    assert mocked_data.call_args_list == [call()]
    assert mocked_table.call_args_list == [call(contents_elt, [{'namespec': 'dummy'}], [{'namespec': 'dummy_proc'}])]
    assert view.view_ctx.process_name == ''
    assert mocked_stats.call_args_list == [call(contents_elt, {})]
    mocker.resetall()
    # test call with process selected but not running on considered node
    # process set in excluded_list
    view.view_ctx.process_name = 'dummy_proc'
    view.view_ctx.get_process_status.return_value = Mock(running_identifiers={'10.0.0.2'})
    view.write_contents(contents_elt)
    assert mocked_data.call_args_list == [call()]
    assert mocked_table.call_args_list == [call(contents_elt, [{'namespec': 'dummy'}], [{'namespec': 'dummy_proc'}])]
    assert view.view_ctx.process_name == ''
    assert mocked_stats.call_args_list == [call(contents_elt, {})]
    mocker.resetall()
    # test call with process selected and running but no statistics on the instance
    view.has_process_statistics = False
    view.view_ctx.process_name = 'dummy'
    view.view_ctx.get_process_status.return_value = Mock(running_identifiers={'10.0.0.1'})
    view.write_contents(contents_elt)
    assert mocked_data.call_args_list == [call()]
    assert mocked_table.call_args_list == [call(contents_elt, [{'namespec': 'dummy'}], [{'namespec': 'dummy_proc'}])]
    assert view.view_ctx.process_name == ''
    assert mocked_stats.call_args_list == [call(contents_elt, {})]
    mocker.resetall()
    # test call with process selected and running, and statistics available
    view.has_process_statistics = True
    view.view_ctx.process_name = 'dummy'
    view.write_contents(contents_elt)
    assert mocked_data.call_args_list == [call()]
    assert mocked_table.call_args_list == [call(contents_elt, [{'namespec': 'dummy_proc'}], [{'namespec': 'dummy'}])]
    assert view.view_ctx.process_name == 'dummy'
    assert mocked_stats.call_args_list == [call(contents_elt, {'namespec': 'dummy'})]


def test_get_process_data(mocker, supvisors, view):
    """ Test the ProcInstanceView.get_process_data method. """
    mocker.patch.object(view, 'sort_data', side_effect=lambda x: (sorted(x, key=lambda y: y['namespec']), []))
    mocked_data = mocker.patch.object(view, 'get_supervisord_data', return_value={'namespec': 'supervisord'})
    # get context
    instance_status = view.sup_ctx.instances['10.0.0.1:25000']
    # test with empty context
    view.view_ctx = Mock(local_identifier='10.0.0.1:25000',
                         **{'get_process_stats.side_effect': [(2, 'stats #1'), (1, None), (4, 'stats #3')]})
    assert view.get_process_data() == ([{'namespec': 'supervisord'}], [])
    assert mocked_data.call_args_list == [call(instance_status)]
    # patch context
    for application_name in ['sample_test_1', 'crash', 'firefox']:
        view.sup_ctx.applications[application_name] = create_application(application_name, supvisors)
    process_data = [('xfontsel', 8, True, False), ('segv', 17, False, False), ('firefox', 26, False, True)]
    for process_name, load, has_crashed, disabled in process_data:
        # create process
        info = process_info_by_name(process_name)
        info['has_crashed'] = has_crashed
        info['disabled'] = disabled
        info['has_stdout'] = True
        info['has_stderr'] = False
        process = create_process(info, supvisors)
        process.rules.expected_load = load
        process.add_info('10.0.0.1:25000', info)
        # add to application
        view.sup_ctx.applications[process.application_name].processes[process.namespec] = process
        # add to supvisors instance status
        instance_status.processes[process.namespec] = process
    # test normal behavior
    sorted_data, excluded_data = view.get_process_data()
    # test intermediate list
    data1 = {'row_type': ProcessRowTypes.INSTANCE_PROCESS,
             'application_name': 'sample_test_1', 'process_name': 'xfontsel', 'namespec': 'sample_test_1:xfontsel',
             'main': False, 'identifier': '10.0.0.1:25000', 'disabled': False,
             'startable': True, 'stoppable': True,
             'statename': 'RUNNING', 'statecode': 20, 'gravity': 'RUNNING', 'has_crashed': True,
             'description': 'pid 80879, uptime 0:01:19',
             'expected_load': 8, 'nb_cores': 2, 'proc_stats': 'stats #1',
             'has_stdout': True, 'has_stderr': False}
    data2 = {'row_type': ProcessRowTypes.INSTANCE_PROCESS,
             'application_name': 'crash', 'process_name': 'segv', 'namespec': 'crash:segv',
             'main': False, 'identifier': '10.0.0.1:25000', 'disabled': False,
             'startable': True, 'stoppable': True,
             'statename': 'BACKOFF', 'statecode': 30, 'gravity': 'BACKOFF', 'has_crashed': False,
             'description': 'Exited too quickly (process log may have details)',
             'expected_load': 17, 'nb_cores': 1, 'proc_stats': None,
             'has_stdout': True, 'has_stderr': False}
    data3 = {'row_type': ProcessRowTypes.INSTANCE_PROCESS,
             'application_name': 'firefox', 'process_name': 'firefox', 'namespec': 'firefox',
             'main': True, 'identifier': '10.0.0.1:25000', 'disabled': True,
             'startable': False, 'stoppable': True,
             'statename': 'EXITED', 'statecode': 100, 'gravity': 'EXITED', 'has_crashed': False,
             'description': 'Sep 14 05:18 PM',
             'expected_load': 26, 'nb_cores': 4, 'proc_stats': 'stats #3',
             'has_stdout': True, 'has_stderr': False}
    assert sorted_data == [data2, data3, data1, {'namespec': 'supervisord'}]
    assert excluded_data == []


def test_get_supervisord_data(view):
    """ Test the ProcInstanceView.get_supervisord_data method. """
    view.view_ctx = Mock(local_identifier='10.0.0.1:25000', **{'get_process_stats.return_value': (2, 'stats #1')})
    # get context
    instance_status = view.sup_ctx.instances['10.0.0.1:25000']
    instance_status.times.start_local_mtime = 0
    pid = os.getpid()
    # test call on empty time values
    supervisord_info = {'row_type': ProcessRowTypes.SUPERVISOR_PROCESS,
                        'application_name': 'supervisord', 'process_name': 'supervisord', 'namespec': 'supervisord',
                        'main': True, 'identifier': '10.0.0.1:25000', 'disabled': False,
                        'startable': False, 'stoppable': True,
                        'description': f'pid {pid}, uptime 0:00:00',
                        'statecode': 20, 'statename': 'RUNNING', 'gravity': 'RUNNING', 'has_crashed': False,
                        'expected_load': 0, 'nb_cores': 2, 'proc_stats': 'stats #1'}
    assert view.get_supervisord_data(instance_status) == supervisord_info
    # test call on relevant time values
    instance_status.times.start_local_mtime = 1000
    instance_status.times.local_mtime = 185618
    supervisord_info = {'row_type': ProcessRowTypes.SUPERVISOR_PROCESS,
                        'application_name': 'supervisord', 'process_name': 'supervisord', 'namespec': 'supervisord',
                        'main': True, 'identifier': '10.0.0.1:25000', 'disabled': False,
                        'startable': False, 'stoppable': True,
                        'description': f'pid {pid}, uptime 2 days, 3:16:58',
                        'statecode': 20, 'statename': 'RUNNING', 'gravity': 'RUNNING', 'has_crashed': False,
                        'expected_load': 0, 'nb_cores': 2, 'proc_stats': 'stats #1'}
    assert view.get_supervisord_data(instance_status) == supervisord_info


def test_sort_data(mocker, view):
    """ Test the ProcInstanceView.sort_data method. """
    mocker.patch.object(view, 'get_application_summary',
                        side_effect=[{'application_name': 'crash', 'process_name': None},
                                     {'application_name': 'sample_test_1', 'process_name': None},
                                     {'application_name': 'sample_test_2', 'process_name': None}] * 2)
    view.view_ctx = Mock(local_identifier='10.0.0.1', **{'get_process_stats.return_value': (2, 'stats #1')})
    # build process list
    processes = [{'application_name': info['group'], 'process_name': info['name'],
                  'main': info['group'] == info['name']}
                 for info in ProcessInfoDatabase]
    shuffle(processes)
    # patch context
    view.view_ctx.get_application_shex.side_effect = [(True, 0), (True, 0), (True, 0),
                                                      (True, 0), (False, 0), (False, 0)]
    # test ordering
    actual, excluded = view.sort_data(processes)
    assert actual == [{'application_name': 'crash', 'process_name': None},
                      {'application_name': 'crash', 'process_name': 'late_segv', 'main': False},
                      {'application_name': 'crash', 'process_name': 'segv', 'main': False},
                      {'application_name': 'firefox', 'process_name': 'firefox', 'main': True},
                      {'application_name': 'sample_test_1', 'process_name': None},
                      {'application_name': 'sample_test_1', 'process_name': 'xclock', 'main': False},
                      {'application_name': 'sample_test_1', 'process_name': 'xfontsel', 'main': False},
                      {'application_name': 'sample_test_1', 'process_name': 'xlogo', 'main': False},
                      {'application_name': 'sample_test_2', 'process_name': None},
                      {'application_name': 'sample_test_2', 'process_name': 'sleep', 'main': False},
                      {'application_name': 'sample_test_2', 'process_name': 'yeux_00', 'main': False},
                      {'application_name': 'sample_test_2', 'process_name': 'yeux_01', 'main': False}]
    assert excluded == []
    # test with some shex on applications
    actual, excluded = view.sort_data(processes)
    assert actual == [{'application_name': 'crash', 'process_name': None},
                      {'application_name': 'crash', 'process_name': 'late_segv', 'main': False},
                      {'application_name': 'crash', 'process_name': 'segv', 'main': False},
                      {'application_name': 'firefox', 'process_name': 'firefox', 'main': True},
                      {'application_name': 'sample_test_1', 'process_name': None},
                      {'application_name': 'sample_test_2', 'process_name': None}]
    sorted_excluded = sorted(excluded, key=lambda x: x['process_name'])
    assert sorted_excluded == [{'application_name': 'sample_test_2', 'process_name': 'sleep', 'main': False},
                               {'application_name': 'sample_test_1', 'process_name': 'xclock', 'main': False},
                               {'application_name': 'sample_test_1', 'process_name': 'xfontsel', 'main': False},
                               {'application_name': 'sample_test_1', 'process_name': 'xlogo', 'main': False},
                               {'application_name': 'sample_test_2', 'process_name': 'yeux_00', 'main': False},
                               {'application_name': 'sample_test_2', 'process_name': 'yeux_01', 'main': False}]


def test_get_application_summary(mocker, view):
    """ Test the ProcInstanceView.get_application_summary method. """
    mocked_sum = mocker.patch.object(view, 'sum_process_info')
    # patch the context
    view.view_ctx = Mock(local_identifier='10.0.0.1')
    view.sup_ctx.applications['dummy_appli'] = Mock(state=ApplicationStates.RUNNING,
                                                    **{'get_operational_status.return_value': 'good'})
    # prepare parameters
    proc_1, proc_2, proc_3, proc_4 = Mock(), Mock(), Mock(), Mock()
    # test with empty list of processes
    expected = {'row_type': ProcessRowTypes.APPLICATION,
                'application_name': 'dummy_appli', 'process_name': None, 'namespec': None,
                'disabled': False, 'startable': False, 'stoppable': True,
                'identifier': '10.0.0.1', 'statename': 'RUNNING', 'statecode': 2, 'gravity': 'RUNNING',
                'has_crashed': False, 'description': 'good', 'nb_items': 0,
                'expected_load': 0, 'nb_cores': 0, 'proc_stats': None}
    mocked_sum.return_value = 0, 0, None
    assert view.get_application_summary('dummy_appli', []) == expected
    # test with non-running processes
    expected.update({'nb_items': 1})
    assert view.get_application_summary('dummy_appli', [proc_4]) == expected
    # test with a mix of running and non-running processes
    appli_stats = Mock()
    mocked_sum.return_value = 27, 8, appli_stats
    expected.update({'nb_items': 4, 'expected_load': 27, 'nb_cores': 8, 'proc_stats': appli_stats})
    assert view.get_application_summary('dummy_appli', [proc_1, proc_2, proc_3, proc_4]) == expected


def test_sum_process_info():
    """ Test the ProcInstanceView.sum_process_info method. """
    # prepare parameters
    proc_1 = {'statecode': ProcessStates.RUNNING, 'expected_load': 5, 'nb_cores': 8,
              'proc_stats': Mock(cpu=[10], mem=[5])}
    proc_2 = {'statecode': ProcessStates.STARTING, 'expected_load': 15, 'nb_cores': 8,
              'proc_stats': Mock(cpu=[], mem=[])}
    proc_3 = {'statecode': ProcessStates.BACKOFF, 'expected_load': 7, 'nb_cores': 8,
              'proc_stats': Mock(cpu=[8], mem=[22])}
    proc_4 = {'statecode': ProcessStates.FATAL, 'expected_load': 25, 'nb_cores': 8,
              'proc_stats': None}
    # test with empty list of processes
    assert ProcInstanceView.sum_process_info([]) == (0, 0, None)
    # test with non-running processes
    assert ProcInstanceView.sum_process_info([proc_4]) == (0, 0, None)
    # test with a mix of running and non-running processes
    expected_load, nb_cores, appli_stats = ProcInstanceView.sum_process_info([proc_1, proc_2, proc_3, proc_4])
    assert expected_load == 27
    assert nb_cores == 8
    assert appli_stats.cpu == [18]
    assert appli_stats.mem == [27]


def test_write_process_table(mocker, view):
    """ Test the ProcInstanceView.write_process_table method. """
    mocked_shex = mocker.patch.object(view, 'write_application_global_shex')
    mocked_total = mocker.patch.object(view, 'write_total_status')
    mocked_appli = mocker.patch.object(view, 'write_application_status')
    mocked_common = mocker.patch.object(view, 'write_common_process_status')
    mocked_supervisord = mocker.patch.object(view, 'write_supervisord_status')
    # patch the meld elements
    tr_elt_0 = Mock(attrib={'class': ''}, **{'findmeld.return_value': Mock()})
    tr_elt_1 = Mock(attrib={'class': ''}, **{'findmeld.return_value': Mock()})
    tr_elt_2 = Mock(attrib={'class': ''}, **{'findmeld.return_value': Mock()})
    tr_elt_3 = Mock(attrib={'class': ''}, **{'findmeld.return_value': Mock()})
    tr_elt_4 = Mock(attrib={'class': ''}, **{'findmeld.return_value': Mock()})
    tr_elt_5 = Mock(attrib={'class': ''}, **{'findmeld.return_value': Mock()})
    tr_mid = Mock(**{'repeat.return_value': [(tr_elt_0, {'row_type': ProcessRowTypes.INSTANCE_PROCESS,
                                                         'process_name': 'info_0', 'main': True}),
                                             (tr_elt_1, {'row_type': ProcessRowTypes.APPLICATION,
                                                         'process_name': None}),
                                             (tr_elt_2, {'row_type': ProcessRowTypes.INSTANCE_PROCESS,
                                                         'process_name': 'info_2', 'main': False}),
                                             (tr_elt_3, {'row_type': ProcessRowTypes.INSTANCE_PROCESS,
                                                         'process_name': 'info_3', 'main': False}),
                                             (tr_elt_4, {'row_type': ProcessRowTypes.APPLICATION,
                                                         'process_name': None}),
                                             (tr_elt_5, {'row_type': ProcessRowTypes.SUPERVISOR_PROCESS,
                                                         'process_name': 'supervisord', 'main': True})]})
    table_mid = Mock(**{'findmeld.return_value': tr_mid})
    mocked_root = Mock(**{'findmeld.return_value': table_mid})
    # test call with no data
    sorted_data, excluded_data = Mock(), Mock()
    view.write_process_table(mocked_root, [], excluded_data)
    assert table_mid.replace.call_args_list == [call('No programs to display')]
    assert not mocked_shex.called
    assert not mocked_common.called
    assert not mocked_total.called
    assert not mocked_appli.called
    assert not tr_elt_0.findmeld.return_value.replace.called
    assert not tr_elt_1.findmeld.return_value.replace.called
    assert not tr_elt_2.findmeld.return_value.replace.called
    assert not tr_elt_3.findmeld.return_value.replace.called
    assert not tr_elt_4.findmeld.return_value.replace.called
    assert not tr_elt_5.findmeld.return_value.replace.called
    assert tr_elt_0.attrib['class'] == ''
    assert tr_elt_1.attrib['class'] == ''
    assert tr_elt_2.attrib['class'] == ''
    assert tr_elt_3.attrib['class'] == ''
    assert tr_elt_4.attrib['class'] == ''
    assert tr_elt_5.attrib['class'] == ''
    table_mid.replace.reset_mock()
    # test call with data and line selected
    view.write_process_table(mocked_root, sorted_data, excluded_data)
    assert not table_mid.replace.called
    assert mocked_shex.call_args_list == [call(table_mid)]
    assert mocked_common.call_args_list == [call(tr_elt_0, {'row_type': ProcessRowTypes.INSTANCE_PROCESS,
                                                            'process_name': 'info_0', 'main': True}),
                                            call(tr_elt_2, {'row_type': ProcessRowTypes.INSTANCE_PROCESS,
                                                            'process_name': 'info_2', 'main': False}),
                                            call(tr_elt_3, {'row_type': ProcessRowTypes.INSTANCE_PROCESS,
                                                            'process_name': 'info_3', 'main': False})]
    assert mocked_supervisord.call_args_list == [call(tr_elt_5, {'row_type': ProcessRowTypes.SUPERVISOR_PROCESS,
                                                                 'process_name': 'supervisord', 'main': True})]
    assert mocked_appli.call_args_list == [call(tr_elt_1, {'row_type': ProcessRowTypes.APPLICATION,
                                                           'process_name': None}, True),
                                           call(tr_elt_4, {'row_type': ProcessRowTypes.APPLICATION,
                                                           'process_name': None}, False)]
    assert not tr_elt_0.findmeld.return_value.replace.called
    assert not tr_elt_1.findmeld.return_value.replace.called
    assert tr_elt_2.findmeld.return_value.replace.call_args_list == [call('')]
    assert tr_elt_3.findmeld.return_value.replace.call_args_list == [call('')]
    assert not tr_elt_4.findmeld.return_value.replace.called
    assert not tr_elt_5.findmeld.return_value.replace.called
    assert tr_elt_0.attrib['class'] == 'brightened'
    assert tr_elt_1.attrib['class'] == 'shaded'
    assert tr_elt_2.attrib['class'] == 'brightened'
    assert tr_elt_3.attrib['class'] == 'shaded'
    assert tr_elt_4.attrib['class'] == 'brightened'
    assert tr_elt_5.attrib['class'] == 'shaded'
    assert mocked_total.call_args_list == [call(table_mid, sorted_data, excluded_data)]


def test_write_application_global_shex(mocker, view):
    """ Test the ProcInstanceView.write_application_global_shex method. """
    mocked_shex = mocker.patch.object(view, 'write_global_shex')
    # add context
    expanded = bytearray.fromhex('ffff')
    shrank = bytearray.fromhex('0000')
    view.view_ctx = Mock(application_shex='1234',
                         **{'get_default_application_shex.side_effect': lambda x: expanded if x else shrank})
    # build XHTML structure
    table_elt = create_element()
    # test call
    view.write_application_global_shex(table_elt)
    assert mocked_shex.call_args_list == [call(table_elt, APP_SHRINK_EXPAND, '1234', expanded, shrank)]


def test_write_application_status(mocker, view):
    """ Test the ProcInstanceView.write_application_status method. """
    mocked_common = mocker.patch.object(view, 'write_common_statistics')
    mocker.patch.object(view, '_write_process_button')
    # patch the context
    view.view_ctx = Mock(**{'get_application_shex.side_effect': [(False, '010'), (True, '101')],
                            'format_url.return_value': 'an url'})
    # patch the meld elements
    shex_a_mid = create_element()
    shex_td_mid = create_element({'shex_a_mid': shex_a_mid})
    name_a_mid = create_element()
    name_td_mid = create_element({'name_a_mid': name_a_mid})
    state_td_mid = create_element()
    desc_td_mid = create_element()
    clear_td_mid = create_element()
    tailout_td_mid = create_element()
    tailerr_td_mid = create_element()
    mocked_root = create_element({'shex_td_mid': shex_td_mid, 'name_td_mid': name_td_mid,
                                  'state_td_mid': state_td_mid, 'desc_td_mid': desc_td_mid,
                                  'clear_td_mid': clear_td_mid,
                                  'tailout_td_mid': tailout_td_mid, 'tailerr_td_mid': tailerr_td_mid})
    # prepare parameters
    info = {'application_name': 'dummy_appli', 'nb_items': 4}
    # test call with application processes hidden
    view.write_application_status(mocked_root, info, True)
    assert mocked_common.call_args_list == [call(mocked_root, info)]
    assert 'rowspan' not in shex_td_mid.attrib
    assert shex_td_mid.attrib['class'] == ''
    assert shex_a_mid.content.call_args_list == [call('[+]')]
    assert shex_a_mid.attributes.call_args_list == [call(href='an url')]
    assert view.view_ctx.format_url.call_args_list == [call('', 'proc_instance.html', ashex='010'),
                                                       call('', 'application.html', appname='dummy_appli')]
    assert name_a_mid.content.call_args_list == [call('dummy_appli')]
    assert name_a_mid.attributes.call_args_list == [call(href='an url')]
    assert clear_td_mid.attrib['colspan'] == '3'
    assert clear_td_mid.content.call_args_list == [call('')]
    for mid in [tailout_td_mid, tailerr_td_mid]:
        assert mid.replace.call_args_list == [call('')]
    # reset context
    mocker.resetall()
    mocked_root.reset_all()
    view.view_ctx.format_url.reset_mock()
    # test call with application processes displayed
    view.write_application_status(mocked_root, info, False)
    assert mocked_common.call_args_list == [call(mocked_root, info)]
    assert shex_td_mid.attrib['rowspan'] == '5'
    assert shex_td_mid.attrib['class'] == 'brightened'
    assert shex_a_mid.content.call_args_list == [call('[\u2013]')]
    assert shex_a_mid.attributes.call_args_list == [call(href='an url')]
    assert view.view_ctx.format_url.call_args_list == [call('', 'proc_instance.html', ashex='101'),
                                                       call('', 'application.html', appname='dummy_appli')]
    assert name_a_mid.content.call_args_list == [call('dummy_appli')]
    assert name_a_mid.attributes.call_args_list == [call(href='an url')]
    assert clear_td_mid.attrib['colspan'] == '3'
    assert clear_td_mid.content.call_args_list == [call('')]
    for mid in [tailout_td_mid, tailerr_td_mid]:
        assert mid.replace.call_args_list == [call('')]


def test_write_supervisord_status(mocker, view):
    """ Test the write_supervisord_status method. """
    mocked_button = mocker.patch.object(view, '_write_supervisord_button')
    mocked_off = mocker.patch.object(view, '_write_supervisord_off_button')
    mocked_state = mocker.patch.object(view, 'write_common_state')
    mocked_stats = mocker.patch.object(view, 'write_common_statistics')
    # patch the view context
    view.view_ctx = Mock(**{'format_url.return_value': 'an url'})
    # patch the meld elements
    shex_elt = create_element()
    name_elt = create_element()
    start_elt = create_element()
    tailerr_elt = create_element()
    tr_elt = create_element({'shex_td_mid': shex_elt, 'name_a_mid': name_elt, 'start_a_mid': start_elt,
                             'tailerr_a_mid': tailerr_elt})
    # test call while not Master
    assert not view.sup_ctx.is_master
    info = {'namespec': 'supervisord', 'process_name': 'supervisord'}
    view.write_supervisord_status(tr_elt, info)
    assert mocked_state.call_args_list == [call(tr_elt, info)]
    assert mocked_stats.call_args_list == [call(tr_elt, info)]
    assert tr_elt.findmeld.call_args_list == [call('name_a_mid')]
    assert not shex_elt.content.called
    assert name_elt.content.call_args_list == [call('supervisord')]
    assert view.view_ctx.format_url.call_args_list == [call('', 'maintail.html', processname='supervisord', limit=1024)]
    assert name_elt.attributes.call_args_list == [call(href='an url', target="_blank")]
    assert mocked_button.call_args_list == [call(tr_elt, 'stop_a_mid', 'proc_instance.html', action='shutdownsup'),
                                            call(tr_elt, 'restart_a_mid', 'proc_instance.html', action='restartsup'),
                                            call(tr_elt, 'clear_a_mid', 'proc_instance.html', action='mainclearlog'),
                                            call(tr_elt, 'tailout_a_mid', MAIN_STDOUT_PAGE)]
    assert mocked_off.call_args_list == [call(tr_elt, 'start_a_mid'), call(tr_elt, 'tailerr_a_mid')]
    mocker.resetall()
    tr_elt.reset_all()
    view.view_ctx.format_url.reset_mock()
    # test call while Master
    view.sup_ctx.master_identifier = view.sup_ctx.local_identifier
    assert view.sup_ctx.is_master
    info = {'namespec': 'supervisord', 'process_name': 'supervisord'}
    view.write_supervisord_status(tr_elt, info)
    assert mocked_state.call_args_list == [call(tr_elt, info)]
    assert mocked_stats.call_args_list == [call(tr_elt, info)]
    assert tr_elt.findmeld.call_args_list == [call('shex_td_mid'), call('name_a_mid')]
    assert shex_elt.content.call_args_list == [call(MASTER_SYMBOL)]
    assert name_elt.content.call_args_list == [call('supervisord')]
    assert view.view_ctx.format_url.call_args_list == [call('', 'maintail.html', processname='supervisord', limit=1024)]
    assert name_elt.attributes.call_args_list == [call(href='an url', target="_blank")]
    assert mocked_button.call_args_list == [call(tr_elt, 'stop_a_mid', 'proc_instance.html', action='shutdownsup'),
                                            call(tr_elt, 'restart_a_mid', 'proc_instance.html', action='restartsup'),
                                            call(tr_elt, 'clear_a_mid', 'proc_instance.html', action='mainclearlog'),
                                            call(tr_elt, 'tailout_a_mid', MAIN_STDOUT_PAGE)]
    assert mocked_off.call_args_list == [call(tr_elt, 'start_a_mid'), call(tr_elt, 'tailerr_a_mid')]


def test_write_supervisord_button(view):
    """ Test the ProcInstanceView._write_supervisord_button method. """
    # patch the view context
    view.view_ctx = Mock(**{'format_url.return_value': 'an url'})
    # patch the meld elements
    a_elt = Mock(attrib={'class': ''})
    tr_elt = Mock(attrib={}, **{'findmeld.return_value': a_elt})
    # test call with action parameters
    view._write_supervisord_button(tr_elt, 'any_a_mid', 'proc_instance.html', **{ACTION: 'any_action'})
    assert tr_elt.findmeld.call_args_list == [call('any_a_mid')]
    assert view.view_ctx.format_url.call_args_list == [call('', 'proc_instance.html', action='any_action')]
    assert a_elt.attrib == {'class': 'button on'}
    assert a_elt.attributes.call_args_list == [call(href='an url')]
    tr_elt.findmeld.reset_mock()
    view.view_ctx.format_url.reset_mock()
    a_elt.attributes.reset_mock()
    a_elt.attrib['class'] = 'active'
    # test call without action parameters
    view._write_supervisord_button(tr_elt, 'any_a_mid', 'proc_instance.html')
    assert tr_elt.findmeld.call_args_list == [call('any_a_mid')]
    assert view.view_ctx.format_url.call_args_list == [call('', 'proc_instance.html')]
    assert a_elt.attrib == {'class': 'active button on'}
    assert a_elt.attributes.call_args_list == [call(href='an url')]


def test_write_supervisord_off_button(view):
    """ Test the ProcInstanceView._write_supervisord_off_button method. """
    # patch the meld elements
    start_a_mid = create_element()
    tr_elt = create_element({'start_a_mid': start_a_mid})
    # test call
    view._write_supervisord_off_button(tr_elt, 'start_a_mid')
    assert tr_elt.findmeld.call_args_list == [call('start_a_mid')]
    assert start_a_mid.attrib == {'class': 'button off'}
    assert not start_a_mid.attributes.called


def test_write_total_status(mocker, supvisors, view):
    """ Test the ProcInstanceView.write_total_status method. """
    mocked_sum = mocker.patch.object(view, 'sum_process_info', return_value=(50, 2, None))
    # patch the meld elements
    load_elt = create_element()
    mem_elt = create_element()
    cpu_elt = create_element()
    tr_elt = create_element({'load_total_th_mid': load_elt, 'mem_total_th_mid': mem_elt, 'cpu_total_th_mid': cpu_elt})
    root_elt = create_element({'total_mid': tr_elt})
    # test call
    sorted_data = [1, 2]
    excluded_data = [3, 4]
    view.write_total_status(root_elt, sorted_data, excluded_data)
    assert mocked_sum.call_args_list == [call([1, 2, 3, 4])]
    assert root_elt.findmeld.call_args_list == [call('total_mid')]
    assert tr_elt.findmeld.call_args_list == [call('load_total_th_mid')]
    assert load_elt.content.call_args_list == [call('50')]
    assert not mem_elt.content.called
    assert not cpu_elt.content.called
    mocked_sum.reset_mock()
    root_elt.findmeld.reset_mock()
    tr_elt.findmeld.reset_mock()
    load_elt.content.reset_mock()
    # test call with process stats and irix mode
    mocked_sum.return_value = 50, 2, Mock(cpu=[12], mem=[25])
    supvisors.options.stats_irix_mode = True
    view.write_total_status(root_elt, sorted_data, excluded_data)
    assert mocked_sum.call_args_list == [call([1, 2, 3, 4])]
    assert root_elt.findmeld.call_args_list == [call('total_mid')]
    assert tr_elt.findmeld.call_args_list == [call('load_total_th_mid'), call('mem_total_th_mid'),
                                              call('cpu_total_th_mid')]
    assert load_elt.content.call_args_list == [call('50')]
    assert mem_elt.content.call_args_list == [call('25.00')]
    assert cpu_elt.content.call_args_list == [call('12.00')]
    mocked_sum.reset_mock()
    root_elt.findmeld.reset_mock()
    tr_elt.findmeld.reset_mock()
    load_elt.content.reset_mock()
    mem_elt.content.reset_mock()
    cpu_elt.content.reset_mock()
    # test call with process stats and solaris mode
    supvisors.options.stats_irix_mode = False
    view.write_total_status(root_elt, sorted_data, excluded_data)
    assert mocked_sum.call_args_list == [call([1, 2, 3, 4])]
    assert root_elt.findmeld.call_args_list == [call('total_mid')]
    assert tr_elt.findmeld.call_args_list == [call('load_total_th_mid'), call('mem_total_th_mid'),
                                              call('cpu_total_th_mid')]
    assert load_elt.content.call_args_list == [call('50')]
    assert mem_elt.content.call_args_list == [call('25.00')]
    assert cpu_elt.content.call_args_list == [call('6.00')]


def test_make_callback(mocker, view):
    """ Test the ProcInstanceView.make_callback method. """
    mocker.patch('supvisors.web.webutils.ctime', return_value='19:10:20')
    mocked_start = mocker.patch.object(view, 'start_group_action', return_value='started')
    mocked_stop = mocker.patch.object(view, 'stop_group_action', return_value='stopped')
    mocked_restart = mocker.patch.object(view, 'restart_group_action', return_value='restarted')
    mocked_clear = mocker.patch.object(view, 'clear_log_action', return_value='cleared')
    mocked_parent = mocker.patch('supvisors.web.viewinstance.SupvisorsInstanceView.make_callback',
                                 return_value='default')
    # test startgroup
    assert view.make_callback('namespec', 'startgroup') == 'started'
    assert mocked_start.call_args_list == [call('namespec')]
    assert not mocked_stop.called
    assert not mocked_restart.called
    assert not mocked_clear.called
    assert not mocked_parent.called
    mocked_start.reset_mock()
    # test stopgroup
    assert view.make_callback('namespec', 'stopgroup') == 'stopped'
    assert not mocked_start.called
    assert mocked_stop.call_args_list == [call('namespec')]
    assert not mocked_restart.called
    assert not mocked_clear.called
    assert not mocked_parent.called
    mocked_stop.reset_mock()
    # test restartgroup
    assert view.make_callback('namespec', 'restartgroup') == 'restarted'
    assert not mocked_start.called
    assert not mocked_stop.called
    assert mocked_restart.call_args_list == [call('namespec')]
    assert not mocked_clear.called
    assert not mocked_parent.called
    mocked_restart.reset_mock()
    # test mainclearlog
    assert view.make_callback('namespec', 'mainclearlog') == 'cleared'
    assert not mocked_start.called
    assert not mocked_stop.called
    assert not mocked_restart.called
    assert mocked_clear.call_args_list == [call()]
    assert not mocked_parent.called
    mocked_clear.reset_mock()
    # test other commands
    assert view.make_callback('namespec', 'other') == 'default'
    assert mocked_parent.call_args_list == [call('namespec', 'other')]
    assert not mocked_clear.called
    mocked_parent.reset_mock()
    # test another command returning an error message different from DISABLED code
    mocked_parent.return_value = lambda: 'abnormal termination'
    result = view.make_callback('namespec', 'other')
    assert result() == 'abnormal termination'
    assert mocked_parent.call_args_list == [call('namespec', 'other')]
    assert not mocked_clear.called
    mocked_parent.reset_mock()
    # test another command returning an error message
    mocked_parent.return_value = lambda: 'unexpected rpc fault [103]'
    result = view.make_callback('namespec', 'other')
    assert result() == ('erro', 'Process namespec: disabled at 19:10:20')
    assert mocked_parent.call_args_list == [call('namespec', 'other')]
    assert not mocked_clear.called


def test_start_group_action(mocker, view):
    """ Test the start_group_action method. """
    mocked_action = mocker.patch.object(view, 'supervisor_rpc_action')
    # test without auto-refresh
    view.view_ctx = Mock(auto_refresh=False)
    view.start_group_action('dummy_proc:*')
    assert mocked_action.call_args_list == [call('startProcess', ('dummy_proc:*', True),
                                                 'Group dummy_proc:* started')]
    mocker.resetall()
    # test with auto-refresh
    view.view_ctx.auto_refresh = True
    view.start_group_action('dummy_proc:*')
    assert mocked_action.call_args_list == [call('startProcess', ('dummy_proc:*', False),
                                                 'Group dummy_proc:* started')]


def test_stop_group_action(mocker, view):
    """ Test the stop_group_action method. """
    mocked_action = mocker.patch.object(view, 'supervisor_rpc_action')
    # test without auto-refresh
    view.view_ctx = Mock(auto_refresh=False)
    view.stop_group_action('dummy_proc:*')
    assert mocked_action.call_args_list == [call('stopProcess', ('dummy_proc:*', True),
                                                 'Group dummy_proc:* stopped')]
    mocker.resetall()
    # test with auto-refresh
    view.view_ctx.auto_refresh = True
    view.stop_group_action('dummy_proc:*')
    assert mocked_action.call_args_list == [call('stopProcess', ('dummy_proc:*', False),
                                                 'Group dummy_proc:* stopped')]


def test_restart_group_action(mocker, view):
    """ Test the restart_group_action method. """
    mocked_action = mocker.patch.object(view, 'multicall_rpc_action')
    # test without auto-refresh
    view.view_ctx = Mock(auto_refresh=False)
    view.restart_group_action('dummy_proc:*')
    multicall = [{'methodName': 'supervisor.stopProcess', 'params': ['dummy_proc:*']},
                 {'methodName': 'supervisor.startProcess', 'params': ['dummy_proc:*', True]}]
    assert mocked_action.call_args_list == [call(multicall, 'Group dummy_proc:* restarted')]
    mocker.resetall()
    # test with auto-refresh
    view.view_ctx.auto_refresh = True
    view.restart_group_action('dummy_proc:*')
    multicall[1]['params'][1] = False
    assert mocked_action.call_args_list == [call(multicall, 'Group dummy_proc:* restarted')]


def test_clear_log_action(mocker, view):
    """ Test the ProcInstanceView.clear_log_action method. """
    mocked_action = mocker.patch.object(view, 'supervisor_rpc_action')
    view.clear_log_action()
    assert mocked_action.call_args_list == [call('clearLog', (), 'Log for Supervisor cleared')]



# ======================================================================
# Copyright 2018 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import re
from unittest.mock import call, Mock

import pytest

from supvisors.statscollector import LocalNodeInfo
from supvisors.web.viewcontext import *
from .base import DummyHttpContext

url_attr_template = r'(.+=.+)'


@pytest.fixture
def http_context(supvisors) -> DummyHttpContext:
    """ Fixture for Dummy HTTP Context. """
    http_context = DummyHttpContext('')
    http_context.supervisord.supvisors = supvisors
    return http_context


@pytest.fixture
def ctx(http_context) -> ViewContext:
    """ Fixture for the instance to test. """
    return ViewContext(http_context)


def test_init(http_context, supvisors, ctx):
    """ Test the values set at ViewContext construction. """
    assert ctx.http_context is http_context
    assert ctx.supvisors is http_context.supervisord.supvisors
    assert ctx.local_identifier == supvisors.mapper.local_identifier
    assert ctx.parameters == {'ident': '10.0.0.4:25000', 'namespec': None, 'period': 5.0,
                              'appname': None, 'processname': None, 'cpuid': 0,
                              'nic': None, 'auto': False, 'strategy': 'CONFIG', 'ashex': '',
                              'diskstats': 'io', 'partition': None, 'device': None}
    # errors must be set due to dummy values
    assert isinstance(ctx.store_message, tuple)
    assert len(ctx.store_message) == 2
    assert ctx.store_message[0] == 'erro'
    assert not ctx.redirect


def test_properties(ctx):
    """ Test the ViewContext properties. """
    assert ctx.strategy == StartingStrategies.CONFIG
    assert not ctx.auto_refresh
    assert ctx.identifier == '10.0.0.4:25000'
    assert ctx.application_name is None
    assert ctx.process_name is None
    ctx.process_name = 'dummy_process_1'
    assert ctx.process_name == 'dummy_process_1'
    assert ctx.namespec is None
    assert ctx.application_shex == ''
    with pytest.raises(KeyError):
        ctx.process_shex
    assert ctx.period == 5.0
    assert ctx.cpu_id == 0
    assert ctx.nic_name is None
    assert ctx.device is None
    assert ctx.partition is None
    assert ctx.disk_stats == 'io'
    assert ctx.action == 'test'
    assert ctx.message == 'hi chaps'
    assert ctx.gravity == 'none'


def test_set_message_default(ctx):
    """ Test the ViewContext.set_message_default method. """
    assert ctx.message == 'hi chaps'
    assert ctx.gravity == 'none'
    # check that the message is not replaced when there is already one stored
    ctx.set_message_default('beware of the dog', 'Warn')
    assert ctx.message == 'hi chaps'
    assert ctx.gravity == 'none'
    # reset the form and retry
    ctx.http_context.form.update({GRAVITY: '', MESSAGE: ''})
    ctx.set_message_default('beware of the dog', 'Warn')
    assert ctx.message == 'beware of the dog'
    assert ctx.gravity == 'Warn'


def test_url_parameters(ctx):
    """ Test the ViewContext.url_parameters method. """
    # test default
    assert ctx.url_parameters(False) == 'diskstats=io&ident=10.0.0.4%3A25000&period=5.0&strategy=CONFIG'
    assert ctx.url_parameters(True) == 'diskstats=io&ident=10.0.0.4%3A25000&period=5.0&strategy=CONFIG'
    # update internal parameters
    ctx.parameters.update({'processname': 'dummy_proc', 'namespec': 'dummy_ns', 'ident': '10.0.0.1:25000', 'cpuid': 3,
                           'appname': 'dummy_appli', 'period': 8, 'strategy': 'CONFIG', 'ashex': '10101',
                           'nic': 'eth0', 'diskstats': 'usage', 'partition': '/', 'device': 'sda'})
    # test without additional parameters
    # don't reset shex
    url = ctx.url_parameters(False)
    # result depends on dict contents so ordering is unreliable
    regexp = r'&'.join([url_attr_template for _ in range(12)])
    matches = re.match(regexp, url)
    assert matches is not None
    expected = sorted(('processname=dummy_proc', 'namespec=dummy_ns', 'ident=10.0.0.1%3A25000', 'cpuid=3',
                       'nic=eth0', 'appname=dummy_appli', 'period=8', 'strategy=CONFIG', 'ashex=10101',
                       'diskstats=usage', 'partition=/', 'device=sda'))
    assert sorted(matches.groups()) == expected
    # reset shex
    url = ctx.url_parameters(True)
    # result depends on dict contents so ordering is unreliable
    regexp = r'&'.join([url_attr_template for _ in range(11)])
    matches = re.match(regexp, url)
    assert matches is not None
    expected = sorted(('processname=dummy_proc', 'namespec=dummy_ns', 'ident=10.0.0.1%3A25000', 'cpuid=3',
                       'nic=eth0', 'appname=dummy_appli', 'period=8', 'strategy=CONFIG',
                       'diskstats=usage', 'partition=/', 'device=sda'))
    assert sorted(matches.groups()) == expected
    # test with additional parameters
    # don't reset shex
    url = ctx.url_parameters(False, **{'ident': '127.0.0.1:25000', 'nic': 'lo', 'ashex': 'args',
                                       'diskstats': 'io', 'partition': '/boot', 'device': 'sdb'})
    regexp = r'&'.join([url_attr_template for _ in range(12)])
    matches = re.match(regexp, url)
    assert matches is not None
    expected = sorted(('processname=dummy_proc', 'namespec=dummy_ns', 'ident=127.0.0.1%3A25000', 'cpuid=3',
                       'nic=lo', 'ashex=args', 'appname=dummy_appli', 'period=8', 'strategy=CONFIG',
                       'diskstats=io', 'partition=/boot', 'device=sdb'))
    assert sorted(matches.groups()) == expected
    # test with additional parameters
    # reset shex
    url = ctx.url_parameters(True, **{'ident': '127.0.0.1:25000', 'nic': 'lo', 'ashex': 'args'})
    regexp = r'&'.join([url_attr_template for _ in range(11)])
    matches = re.match(regexp, url)
    assert matches is not None
    expected = sorted(('processname=dummy_proc', 'namespec=dummy_ns', 'ident=127.0.0.1%3A25000', 'cpuid=3',
                       'nic=lo', 'appname=dummy_appli', 'period=8', 'strategy=CONFIG',
                       'diskstats=usage', 'partition=/', 'device=sda'))
    assert sorted(matches.groups()) == expected


def test_cpu_id_to_string():
    """ Test the ViewContext.cpu_id_to_string method. """
    for idx in range(1, 10):
        assert ViewContext.cpu_id_to_string(idx) == str(idx - 1)
    assert ViewContext.cpu_id_to_string(0) == 'all'
    assert ViewContext.cpu_id_to_string(-5) == 'all'


def test_update_string(ctx):
    """ Test the ViewContext._update_string method. """
    # keep a copy of parameters
    ref_parameters = ctx.parameters.copy()
    # test with unknown parameter and no default value
    assert 'dummy' not in ctx.http_context.form
    ctx._update_string('dummy', [])
    ref_parameters.update({'dummy': None})
    assert ctx.parameters == ref_parameters
    # test with unknown parameter and default value
    assert 'dummy' not in ctx.http_context.form
    ctx._update_string('dummy', [], 'hello')
    ref_parameters.update({'dummy': 'hello'})
    assert ctx.parameters == ref_parameters
    # test with known parameter, no default value and value out of list
    assert 'action' in ctx.http_context.form
    ctx._update_string('action', [])
    ref_parameters.update({'action': None})
    assert ctx.parameters == ref_parameters
    # test with known parameter, default value and value out of list
    assert 'action' in ctx.http_context.form
    ctx._update_string('action', [], 'do it')
    ref_parameters.update({'action': 'do it'})
    assert ctx.parameters == ref_parameters
    # test with known parameter and value in list
    assert 'action' in ctx.http_context.form
    ctx._update_string('action', ['try', 'do it', 'test'])
    ref_parameters.update({'action': 'test'})
    assert ctx.parameters == ref_parameters


def test_update_integer(ctx):
    """ Test the ViewContext._update_integer method. """
    # keep a copy of parameters
    ref_parameters = ctx.parameters.copy()
    # test with unknown parameter and no default value
    assert 'dummy' not in ctx.http_context.form
    ctx._update_integer('dummy', [])
    ref_parameters.update({'dummy': 0})
    assert ctx.parameters == ref_parameters
    # test with unknown parameter and default value
    assert 'dummy' not in ctx.http_context.form
    ctx._update_integer('dummy', [], 'hello')
    ref_parameters.update({'dummy': 'hello'})
    assert ctx.parameters == ref_parameters
    # test with known parameter, not integer and no default value
    assert 'action' in ctx.http_context.form
    ctx._update_integer('action', [])
    ref_parameters.update({'action': 0})
    # test with known parameter, not integer and default value
    assert 'action' in ctx.http_context.form
    ctx._update_integer('action', [], 5)
    ref_parameters.update({'action': 5})
    # test with known parameter, integer, no default value and value out of list
    assert 'SERVER_PORT' in ctx.http_context.form
    ctx._update_integer('SERVER_PORT', [])
    ref_parameters.update({'SERVER_PORT': 0})
    assert ctx.parameters == ref_parameters
    # test with known parameter, integer, default value and value out of list
    assert 'SERVER_PORT' in ctx.http_context.form
    ctx._update_integer('SERVER_PORT', [], 1234)
    ref_parameters.update({'SERVER_PORT': 1234})
    assert ctx.parameters == ref_parameters
    # test with known parameter, integer and value in list
    assert 'SERVER_PORT' in ctx.http_context.form
    ctx._update_integer('SERVER_PORT', [12, 7777, 654])
    ref_parameters.update({'SERVER_PORT': 7777})
    assert ctx.parameters == ref_parameters


def test_update_float(ctx):
    """ Test the ViewContext._update_float method. """
    # keep a copy of parameters
    ref_parameters = ctx.parameters.copy()
    # test with unknown parameter and no default value
    assert 'dummy' not in ctx.http_context.form
    ctx._update_float('dummy', [])
    ref_parameters.update({'dummy': 0})
    assert ctx.parameters == ref_parameters
    # test with unknown parameter and default value
    assert 'dummy' not in ctx.http_context.form
    ctx._update_float('dummy', [], 'hello')
    ref_parameters.update({'dummy': 'hello'})
    assert ctx.parameters == ref_parameters
    # test with known parameter, not float and no default value
    assert 'action' in ctx.http_context.form
    ctx._update_float('action', [])
    ref_parameters.update({'action': 0.0})
    # test with known parameter, not float and default value
    assert 'action' in ctx.http_context.form
    ctx._update_float('action', [], 5.0)
    ref_parameters.update({'action': 5.0})
    # test with known parameter, float, no default value and value out of list
    assert 'period' in ctx.http_context.form
    ctx._update_float('period', [])
    ref_parameters.update({'period': 0.0})
    assert ctx.parameters == ref_parameters
    # test with known parameter, float, default value and value out of list
    assert 'period' in ctx.http_context.form
    ctx._update_float('period', [], 12.3)
    ref_parameters.update({'period': 12.3})
    assert ctx.parameters == ref_parameters
    # test with known parameter, float and value in list
    assert 'period' in ctx.http_context.form
    ctx._update_float('period', [5.1, 15.0, 60.0])
    ref_parameters.update({'period': 5.1})
    assert ctx.parameters == ref_parameters


def test_update_boolean(ctx):
    """ Test the ViewContext._update_boolean method. """
    # keep a copy of parameters
    ref_parameters = ctx.parameters.copy()
    # test with unknown parameter and no default value
    assert 'dummy' not in ctx.http_context.form
    ctx._update_boolean('dummy')
    ref_parameters.update({'dummy': False})
    assert ctx.parameters == ref_parameters
    # test with unknown parameter and default value
    assert 'dummy' not in ctx.http_context.form
    ctx._update_boolean('dummy', True)
    ref_parameters.update({'dummy': True})
    assert ctx.parameters == ref_parameters
    # test with known parameter, no default value and unexpected value
    ctx.http_context.form['auto'] = 'unexpected value'
    ctx._update_boolean('auto')
    ref_parameters.update({'auto': False})
    assert ctx.parameters == ref_parameters
    # test with known parameter, default value and unexpected value
    ctx._update_boolean('auto', True)
    ref_parameters.update({'auto': True})
    assert ctx.parameters == ref_parameters
    # test with known parameter and expected value
    ctx.http_context.form['auto'] = '1'
    ctx._update_boolean('auto')
    ref_parameters.update({'auto': True})
    assert ctx.parameters == ref_parameters


def test_update_period(mocker, ctx):
    """ Test the ViewContext.update_period method. """
    mocked_update = mocker.patch('supvisors.web.viewcontext.ViewContext._update_float')
    # test method with statistics enabled
    ctx.update_period()
    assert mocked_update.call_args_list == [call(PERIOD, ctx.supvisors.options.stats_periods,
                                                 ctx.supvisors.options.stats_periods[0])]
    mocker.resetall()
    # test method with statistics disabled
    ctx.supvisors.options.stats_periods = []
    with pytest.raises(StopIteration):
        ctx.update_period()
    assert not mocked_update.called


def test_update_identifier(ctx):
    """ Test the ViewContext.update_identifier method. """
    # reset parameter because called in constructor
    del ctx.parameters[IDENTIFIER]
    # test call with valid value
    ctx.update_identifier()
    assert ctx.parameters[IDENTIFIER] == '10.0.0.4:25000'
    # reset parameter
    del ctx.parameters[IDENTIFIER]
    # test call with invalid value
    ctx.http_context.form[IDENTIFIER] = '192.168.1.1:25000'
    ctx.update_identifier()
    assert ctx.parameters[IDENTIFIER] == ctx.local_identifier


def test_update_auto_refresh(ctx):
    """ Test the ViewContext.update_auto_refresh method. """
    # reset parameter because called in constructor
    del ctx.parameters[AUTO]
    # test call with default valid value
    ctx.update_auto_refresh()
    assert not ctx.parameters[AUTO]
    # reset parameter
    del ctx.parameters[AUTO]
    # test call with other valid value
    ctx.http_context.form[AUTO] = 't'
    ctx.update_auto_refresh()
    assert ctx.parameters[AUTO]
    # reset parameter
    del ctx.parameters[AUTO]
    # test call with invalid value
    ctx.http_context.form[AUTO] = 'not a boolean'
    ctx.update_auto_refresh()
    assert not ctx.parameters[AUTO]


def test_update_application_name(ctx):
    """ Test the ViewContext.update_application_name method. """
    # reset parameter because called in constructor
    del ctx.parameters[APPLI]
    # test call with valid value
    ctx.supvisors.context.applications = {'abc': [], 'dummy_appli': []}
    ctx.update_application_name()
    # cannot rely on ordering for second parameter because of dict need to split checking
    assert ctx.parameters[APPLI] == 'dummy_appli'
    # reset parameter
    del ctx.parameters[APPLI]
    # test call with invalid value
    ctx.http_context.form[APPLI] = 'any_appli'
    ctx.update_application_name()
    assert ctx.parameters[APPLI] is None


def test_update_process_name(mocker, ctx):
    """ Test the ViewContext.update_process_name method. """
    ctx.parameters[IDENTIFIER] = ctx.local_identifier
    node = ctx.supvisors.context.instances[ctx.local_identifier]
    mocker.patch.object(node, 'running_processes', return_value=[])
    # reset parameter because called in constructor
    del ctx.parameters[PROCESS]
    # test call in case of node stats are not found
    ctx.update_process_name()
    assert ctx.parameters[PROCESS] is None
    # reset parameter
    del ctx.parameters[PROCESS]
    # test call when address stats are found and process in list
    node.running_processes.return_value = [Mock(namespec='abc'), Mock(namespec='dummy_proc')]
    ctx.update_process_name()
    assert ctx.parameters[PROCESS] == 'dummy_proc'
    # reset parameter
    del ctx.parameters[PROCESS]
    # test call when address stats are found and process not in list
    ctx.http_context.form[PROCESS] = 'any_proc'
    ctx.update_process_name()
    assert ctx.parameters[PROCESS] is None


def test_update_namespec(mocker, ctx):
    """ Test the ViewContext.update_namespec method. """
    # reset parameter because called in constructor
    del ctx.parameters[NAMESPEC]
    # test call with no value
    ctx.update_namespec()
    assert ctx.parameters[NAMESPEC] is None
    # test call with invalid value
    ctx.http_context.form[NAMESPEC] = 'any_proc'
    ctx.update_namespec()
    assert ctx.parameters[NAMESPEC] is None
    # test call with valid parameter
    mocker.patch.object(ctx.http_context.supervisord.supvisors.context, 'is_namespec', return_value=True)
    ctx.update_namespec()
    assert ctx.parameters[NAMESPEC] == 'any_proc'


def test_update_cpu_id(mocker, ctx):
    """ Test the ViewContext.update_cpu_id method. """
    mocker.patch('supvisors.web.viewcontext.ViewContext.get_nb_cores', return_value=2)
    mocked_update = mocker.patch('supvisors.web.viewcontext.ViewContext._update_integer')
    # test call
    ctx.update_cpu_id()
    assert mocked_update.call_args_list == [call(CPU, [0, 1, 2])]


def test_update_nic_name(mocker, ctx):
    """ Test the ViewContext.update_nic_name method. """
    mocked_stats = mocker.patch.object(ctx, 'get_instance_stats', return_value=None)
    # reset parameter because called in constructor
    del ctx.parameters[NIC]
    # test call in case of host stats are not found
    ctx.update_nic_name()
    assert ctx.parameters[NIC] is None
    # reset parameter
    del ctx.parameters[NIC]
    # test call when host stats are found and network interface not in list
    mocked_stats.return_value = Mock(net_io={'lo': None})
    ctx.update_nic_name()
    assert ctx.parameters[NIC] == 'lo'
    # reset parameter
    del ctx.parameters[NIC]
    # test call when host stats are found and network interface  in list
    mocked_stats.return_value = Mock(net_io={'lo': None, 'eth0': None})
    ctx.update_nic_name()
    assert ctx.parameters[NIC] == 'eth0'


def test_update_disk_stats_choice(ctx):
    """ Test the ViewContext.update_disk_stats_choice method. """
    del ctx.parameters[DISK_STATS]
    ctx.update_disk_stats_choice()
    assert ctx.parameters[DISK_STATS] == 'io'


def test_update_partition_name(mocker, ctx):
    """ Test the ViewContext.update_partition_name method. """
    mocked_stats = mocker.patch.object(ctx, 'get_instance_stats', return_value=None)
    # reset parameter because called in constructor
    del ctx.parameters[PARTITION]
    # test call in case of host stats are not found
    ctx.update_partition_name()
    assert ctx.parameters[PARTITION] is None
    # reset parameter
    del ctx.parameters[PARTITION]
    # test call when host stats are found and process not in list
    mocked_stats.return_value = Mock(disk_usage={'/boot': None})
    ctx.update_partition_name()
    assert ctx.parameters[PARTITION] == '/boot'
    # reset parameter
    del ctx.parameters[PARTITION]
    # test call when address stats are found and process not in list
    mocked_stats.return_value = Mock(disk_usage={'/': None, '/boot': None})
    ctx.update_partition_name()
    assert ctx.parameters[PARTITION] == '/'


def test_update_device_name(mocker, ctx):
    """ Test the ViewContext.update_device_name method. """
    mocked_stats = mocker.patch.object(ctx, 'get_instance_stats', return_value=None)
    # reset parameter because called in constructor
    del ctx.parameters[DEVICE]
    # test call in case of host stats are not found
    ctx.update_device_name()
    assert ctx.parameters[DEVICE] is None
    # reset parameter
    del ctx.parameters[DEVICE]
    # test call when host stats are found and process in list
    mocked_stats.return_value = Mock(disk_io={'sdb': None})
    ctx.update_device_name()
    assert ctx.parameters[DEVICE] == 'sdb'
    # reset parameter
    del ctx.parameters[DEVICE]
    # test call when address stats are found and process not in list
    mocked_stats.return_value = Mock(disk_io={'sdb': None, 'sda': None})
    ctx.update_device_name()
    assert ctx.parameters[DEVICE] == 'sda'


def test_format_url(supvisors, ctx):
    """ Test the ViewContext.format_url method. """
    # test without node and arguments
    expected = 'index.html?diskstats=io&ident=10.0.0.4%3A25000&period=5.0&strategy=CONFIG'
    assert ctx.format_url(None, 'index.html') == expected
    # test with local node and arguments
    local_instance = supvisors.mapper.local_instance
    base_address = f'http://{local_instance.host_id}:25000/index.html?'
    url = ctx.format_url(ctx.local_identifier, 'index.html',
                         **{'period': 10, 'appliname': 'dummy_appli', 'ashex': 'args'})
    expected = 'appliname=dummy_appli&ashex=args&diskstats=io&ident=10.0.0.4%3A25000&period=10&strategy=CONFIG'
    assert url == base_address + expected
    # test with remote node and arguments (shex expected to be removed)
    url = ctx.format_url('10.0.0.1:25000', 'index.html',
                         **{'period': 10, 'appliname': 'dummy_appli', 'ashex': 'args'})
    base_address = 'http://10.0.0.1:25000/index.html?'
    expected = 'appliname=dummy_appli&diskstats=io&ident=10.0.0.4%3A25000&period=10&strategy=CONFIG'
    assert url == base_address + expected


def test_fire_message(ctx):
    """ Test the ViewContext.fire_message method. """
    ctx.store_message = ('warning', 'not as expected')
    ctx.fire_message()
    # result depends on dict contents so ordering is unreliable
    url = ctx.http_context.response['headers']['Location']
    base_address = 'http://10.0.0.1:7777/index.html?'
    expected = ('diskstats=io&gravity=warning&ident=10.0.0.4%3A25000&message=not%20as%20expected&period=5.0'
                '&strategy=CONFIG')
    assert url == base_address + expected


def test_get_nb_cores(ctx):
    """ Test the ViewContext.get_nb_cores method. """
    # test default
    assert ctx.get_nb_cores(ctx.local_identifier) == 0
    # mock the structure
    stats = ctx.http_context.supervisord.supvisors.host_compiler
    stats.nb_cores[ctx.local_identifier] = 4
    # test new call
    assert ctx.get_nb_cores(ctx.local_identifier) == 4
    # test with unknown address
    assert ctx.get_nb_cores('10.0.0.1:25000') == 0
    # test with known address
    stats.nb_cores['10.0.0.1:25000'] = 8
    assert ctx.get_nb_cores('10.0.0.1:25000') == 8


def test_get_node_characteristics(ctx):
    """ Test the ViewContext.get_node_characteristics method. """
    # test with stats collector present
    assert ctx.supvisors.stats_collector
    assert type(ctx.get_node_characteristics()) is LocalNodeInfo
    # cannot test the LocalNodeInfo contents as it is platform-dependent
    # test with stats collector not set
    ctx.supvisors.stats_collector = None
    assert ctx.get_node_characteristics() is None


def test_get_node_stats(supvisors, ctx):
    """ Test the ViewContext.get_instance_stats method. """
    # test with default address
    instance_stats = ctx.get_instance_stats()
    assert instance_stats.identifier == supvisors.mapper.local_identifier
    assert instance_stats.period == 5.0
    # test with unknown identifier
    assert ctx.get_instance_stats('10.0.0.0:25000') is None
    # test with known address parameter and existing period
    instance_stats = ctx.get_instance_stats('10.0.0.1:25000')
    assert instance_stats.identifier == '10.0.0.1:25000'
    assert instance_stats.period == 5.0
    # update with unknown period
    ctx.parameters[PERIOD] = 8
    assert ctx.get_instance_stats('10.0.0.1:25000') is None


def test_get_process_stats(mocker, supvisors, ctx):
    """ Test the ViewContext.get_process_stats method. """
    mocked_core = mocker.patch('supvisors.web.viewcontext.ViewContext.get_nb_cores', return_value=4)
    # test no result as no data stored
    assert ctx.get_process_stats('dummy_proc', ctx.local_identifier) == (4, None)
    mocked_core.reset_mock()
    # fill some internal structures
    mocked_stats = mocker.patch.object(supvisors.process_compiler, 'get_stats', return_value='mock stats')
    assert ctx.get_process_stats('dummy_proc', '10.0.0.1') == (4, 'mock stats')
    assert mocked_core.call_args_list == [call('10.0.0.1')]
    assert mocked_stats.call_args_list == [call('dummy_proc', '10.0.0.1', 5.0)]


def test_get_process_status(mocker, ctx):
    """ Test the ViewContext.get_process_status method. """
    # test with empty context and nothing in http form
    mocked_get = mocker.patch.object(ctx.supvisors.context, 'get_process', side_effect=KeyError)
    assert ctx.get_process_status() is None
    assert not mocked_get.called
    assert ctx.get_process_status('abc') is None
    assert mocked_get.call_args_list == [call('abc')]
    mocked_get.reset_mock()
    # test with empty context and namespec in http form
    ctx.parameters[NAMESPEC] = 'abc'
    assert ctx.get_process_status() is None
    assert mocked_get.call_args_list == [call('abc')]
    mocked_get.reset_mock()
    assert ctx.get_process_status('dummy_proc') is None
    assert mocked_get.call_args_list == [call('dummy_proc')]
    mocked_get.reset_mock()
    # test with context and namespec in http form
    mocked_get.side_effect = None
    mocked_get.return_value = 'dummy_proc'
    assert ctx.get_process_status() == 'dummy_proc'
    assert ctx.get_process_status('abc') == 'dummy_proc'
    # test with context and nothing in http form
    ctx.parameters[NAMESPEC] = None
    assert ctx.get_process_status() is None
    assert ctx.get_process_status('dummy_proc') == 'dummy_proc'


def test_get_default_shex(ctx):
    """ Test the ViewContext._get_default_shex method. """
    assert ctx._get_default_shex(25, True).hex() == 'ffffffff'
    assert ctx._get_default_shex(16, True).hex() == 'ffff'
    assert ctx._get_default_shex(11, False).hex() == '0000'
    assert ctx._get_default_shex(7, False).hex() == '00'


def test_get_default_application_shex(supvisors, ctx):
    """ Test the ViewContext.get_default_application_shex method. """
    supvisors.context.applications = {f'appli_{x}': Mock() for x in range(15)}
    assert ctx.get_default_application_shex(True).hex() == 'ffff'
    assert ctx.get_default_application_shex(False).hex() == '0000'


def test_get_default_process_shex(supvisors, ctx):
    """ Test the ViewContext.get_default_process_shex method. """
    supvisors.context.applications = {'dummy_appli': Mock(processes=list(range(17)))}
    assert ctx.get_default_process_shex('dummy_appli', True).hex() == 'ffffff'
    assert ctx.get_default_process_shex('dummy_appli', False).hex() == '000000'


def test_update_application_shrink_expand(supvisors, ctx):
    """ Test the ViewContext.update_application_shrink_expand method. """
    # check default
    assert ctx.application_shex == ''
    assert APP_SHRINK_EXPAND not in ctx.http_context.form
    # test with applications in the context
    supvisors.context.applications = {'abc': [], 'def': [], 'ghi': []}
    # test with unknown parameter and no default value
    ctx.update_application_shrink_expand()
    assert ctx.application_shex == 'ff'
    # add unexpected value in form (there should be only even number of hexadecimal chars)
    ctx.http_context.form[APP_SHRINK_EXPAND] = '12A'
    ctx.update_application_shrink_expand()
    assert ctx.application_shex == 'ff'
    ctx.http_context.form[APP_SHRINK_EXPAND] = '12AG'
    ctx.update_application_shrink_expand()
    assert ctx.application_shex == 'ff'
    # update form with unexpected value (string length should be equal to the number of applications)
    ctx.http_context.form[APP_SHRINK_EXPAND] = '0101'
    ctx.update_application_shrink_expand()
    assert ctx.application_shex == 'ff'
    # update form with valid value
    ctx.http_context.form[APP_SHRINK_EXPAND] = '9a'
    ctx.update_application_shrink_expand()
    assert ctx.application_shex == '9a'


def test_update_process_shrink_expand(supvisors, ctx):
    """ Test the ViewContext.update_process_shrink_expand method. """
    # check default
    assert not ctx.application_name
    assert PROC_SHRINK_EXPAND not in ctx.parameters
    assert PROC_SHRINK_EXPAND not in ctx.http_context.form
    # test with no application set
    ctx.update_process_shrink_expand()
    assert PROC_SHRINK_EXPAND not in ctx.parameters
    # test with applications in the context, but still no application set in context form
    supvisors.context.applications['dummy_appli'] = Mock(processes={'abc': [], 'def': [], 'ghi': []})
    ctx.update_process_shrink_expand()
    assert PROC_SHRINK_EXPAND not in ctx.parameters
    # set the application in the context form
    ctx.http_context.form[APPLI] = 'dummy_appli'
    ctx.update_application_name()
    # test with unknown parameter and no default value
    ctx.update_process_shrink_expand()
    assert ctx.process_shex == '00'
    # add unexpected value in form (there should be only even number of hexadecimal chars)
    ctx.http_context.form[PROC_SHRINK_EXPAND] = '12A'
    ctx.update_process_shrink_expand()
    assert ctx.process_shex == '00'
    ctx.http_context.form[PROC_SHRINK_EXPAND] = '12AG'
    ctx.update_process_shrink_expand()
    assert ctx.process_shex == '00'
    # update form with unexpected value (string length should be equal to the number of processes)
    ctx.http_context.form[PROC_SHRINK_EXPAND] = '0101'
    ctx.update_process_shrink_expand()
    assert ctx.process_shex == '00'
    # update form with valid value
    ctx.http_context.form[PROC_SHRINK_EXPAND] = '9a'
    ctx.update_process_shrink_expand()
    assert ctx.process_shex == '9a'


def test_get_application_shex(supvisors, ctx):
    """ Test the ViewContext.get_application_shex method. """
    # patch the context
    supvisors.context.applications = {'abc': [], 'def': [], 'ghi': []}
    # only 'def' is visible
    ba = bytearray([0xff])
    set_bit(ba, 0, 0)
    set_bit(ba, 2, 0)
    assert ba.hex() == 'fa'
    ctx.parameters[APP_SHRINK_EXPAND] = 'fa'
    # test calls and inversion
    assert ctx.get_application_shex('abc') == (False, 'fb')
    assert get_bit(ba.fromhex('fb'), 0)
    assert get_bit(ba.fromhex('fb'), 1)
    assert not get_bit(ba.fromhex('fb'), 2)
    assert ctx.get_application_shex('def') == (True, 'f8')
    assert not get_bit(ba.fromhex('f8'), 0)
    assert not get_bit(ba.fromhex('f8'), 1)
    assert not get_bit(ba.fromhex('f8'), 2)
    assert ctx.get_application_shex('ghi') == (False, 'fe')
    assert not get_bit(ba.fromhex('fe'), 0)
    assert get_bit(ba.fromhex('fe'), 1)
    assert get_bit(ba.fromhex('fe'), 2)


def test_get_process_shex(supvisors, ctx):
    """ Test the ViewContext.get_process_shex method. """
    # patch the context
    supvisors.context.applications['dummy_appli'] = Mock(processes={'abc': [], 'def': [], 'ghi': []})
    # set the application in the context form
    ctx.http_context.form[APPLI] = 'dummy_appli'
    ctx.update_application_name()
    # only 'def' is visible
    ba = bytearray([0xff])
    set_bit(ba, 0, 0)
    set_bit(ba, 2, 0)
    assert ba.hex() == 'fa'
    ctx.parameters[PROC_SHRINK_EXPAND] = 'fa'
    # test calls and inversion
    assert ctx.get_process_shex('abc') == (False, 'fb')
    assert get_bit(ba.fromhex('fb'), 0)
    assert get_bit(ba.fromhex('fb'), 1)
    assert not get_bit(ba.fromhex('fb'), 2)
    assert ctx.get_process_shex('def') == (True, 'f8')
    assert not get_bit(ba.fromhex('f8'), 0)
    assert not get_bit(ba.fromhex('f8'), 1)
    assert not get_bit(ba.fromhex('f8'), 2)
    assert ctx.get_process_shex('ghi') == (False, 'fe')
    assert not get_bit(ba.fromhex('fe'), 0)
    assert get_bit(ba.fromhex('fe'), 1)
    assert get_bit(ba.fromhex('fe'), 2)



# ======================================================================
# Copyright 2016 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

from supvisors.ttypes import SupvisorsStates
from .viewcontext import *
from .viewhandler import ViewHandler
from .webutils import *


class MainView(ViewHandler):
    """ Class ensuring the common rendering of the main pages (index and conciliation) with:

        - a navigation menu towards Supvisors instances contents and applications,
        - the state of Supvisors,
        - actions on Supvisors.
    """

    def __init__(self, context):
        """ Call of the superclass constructors. """
        ViewHandler.__init__(self, context)
        # global actions (no parameter)
        self.global_methods = {'sup_restart': self.sup_restart_action,
                               'sup_shutdown': self.sup_shutdown_action}

    def write_navigation(self, root) -> None:
        """ Rendering of the navigation menu. """
        self.write_nav(root, source=self.local_identifier)

    def write_status(self, header_elt) -> None:
        """ Rendering of the header part of the Supvisors main page. """
        # set Supvisors state & modes
        state_modes = self.sup_ctx.get_state_modes()
        elt = header_elt.findmeld('state_a_mid')
        if state_modes['fsm_statecode'] == SupvisorsStates.CONCILIATION.value:
            elt.attributes(href=CONCILIATION_PAGE)
            statename = state_modes['fsm_statename']
            # blinking state until full conciliation performed
            if self.sup_ctx.conflicting():
                statename += ' >>'
                update_attrib(elt, 'class', 'on blink')
            elt.content(statename)
        else:
            elt.replace(state_modes['fsm_statename'])
        # set Supvisors modes
        for mid, attr in [('starting_mid', 'starting_jobs'), ('stopping_mid', 'stopping_jobs')]:
            elt = header_elt.findmeld(mid)
            if state_modes[attr]:
                update_attrib(elt, 'class', 'blink')
            else:
                elt.replace('')
        # write the Master nick identifier
        master_instance = self.sup_ctx.master_instance
        identifier = master_instance.supvisors_id.nick_identifier if master_instance else 'none'
        header_elt.findmeld('master_name_mid').content(identifier)

    def write_actions(self, header_elt) -> None:
        """ Write actions related to Supvisors. """
        super().write_actions(header_elt)
        # configure restart button
        elt = header_elt.findmeld('restart_a_mid')
        url = self.view_ctx.format_url('', SUPVISORS_PAGE, **{ACTION: 'sup_restart'})
        elt.attributes(href=url)
        # configure shutdown button
        elt = header_elt.findmeld('shutdown_a_mid')
        url = self.view_ctx.format_url('', SUPVISORS_PAGE, **{ACTION: 'sup_shutdown'})
        elt.attributes(href=url)

    def make_callback(self, namespec: str, action: str):
        """ Triggers processing iaw action requested. """
        if action in self.global_methods:
            return self.global_methods[action]()

    def sup_restart_action(self):
        """ Restart all Supervisor instances. """
        try:
            self.supvisors.supervisor_data.supvisors_rpc_interface.restart()
        except RPCError as e:
            return delayed_error(f'restart: {e}')
        return delayed_warn('Supvisors restart requested')

    def sup_shutdown_action(self):
        """ Stop all Supervisor instances. """
        try:
            self.supvisors.supervisor_data.supvisors_rpc_interface.shutdown()
        except RPCError as e:
            return delayed_error(f'shutdown: {e}')
        return delayed_warn('Supvisors shutdown requested')
